# -*- coding: utf-8 -*-
"""
Custom QDoubleValidator that pays attention to ranges

https://gist.github.com/jirivrany/4473639

:copyright:
    Mazama Science, IRIS
:license:
    GNU Lesser General Public License, Version 3
    (http://www.gnu.org/copyleft/lesser.html)
"""

from sys import float_info

from PyQt4.QtGui import QValidator
from PyQt4.QtGui import QDoubleValidator


class MyDoubleValidator(QDoubleValidator):
    '''
    Custom QDoubleValidator that pays attention to ranges

    from https://gist.github.com/jirivrany/4473639

    Note that the Validate API is different for QString v2
    '''

    def __init__(self, bottom=float_info.min,
                 top=float_info.max,
                 decimals=float_info.dig, parent=None):

        super(MyDoubleValidator, self).__init__(bottom, top, decimals, parent)

    def validate(self, input_value, pos):

        if input_value in ('', '.', '-'):
            return QValidator.Intermediate, input_value, pos

        state, value, pos = QDoubleValidator.validate(self, input_value, pos)

        if state != QValidator.Acceptable:
            return QValidator.Invalid, value, pos

        return QValidator.Acceptable, value, pos



# -*- coding: utf-8 -*-
"""
Dialog showing logging information.

:copyright:
    Mazama Science, IRIS
:license:
    GNU Lesser General Public License, Version 3
    (http://www.gnu.org/copyleft/lesser.html)
"""

from pyweed.gui.BaseDialog import BaseDialog
from pyweed.gui.uic import LoggingDialog
from pyweed.gui.MyTextEditLoggingHandler import MyTextEditLoggingHandler
import logging
from PyQt4 import QtCore


class LoggingDialog(BaseDialog, LoggingDialog.Ui_LoggingDialog):
    """
    Dialog window displaying all logs.
    """

    append = QtCore.pyqtSignal(str)

    def __init__(self, parent=None):
        super(LoggingDialog, self).__init__(parent=parent)
        self.setupUi(self)
        self.setWindowTitle('Logs')

        # Initialize loggingPlainTextEdit
        self.loggingPlainTextEdit.setReadOnly(True)

        # Add a widget logging handler to the logger
        loggingHandler = MyTextEditLoggingHandler(signal=self.append)
        formatter = logging.Formatter('%(asctime)s - %(levelname)s - %(message)s', datefmt='%Y-%m-%d %H:%M:%S')
        loggingHandler.setFormatter(formatter)

        logging.getLogger().addHandler(loggingHandler)

        self.append.connect(self.appendMessage)

    def appendMessage(self, msg):
        self.loggingPlainTextEdit.appendPlainText(msg)




# -*- coding: utf-8 -*-
"""
Helper class for web services.

Options represents a set of key/value pairs (ie. query parameters), internally these are typed values but
they can be read/written as strings.

>>> class Test(Options):
...     option1 = Option()
...     option2 = DateOption()
>>> t = Test()
>>> t.option1 = 'test'
>>> t.option1
'test'
>>> t.option2 = '2012-01-01'
>>> t.option2
UTCDateTime(2012, 1, 1, 0, 0)
>>> t.get_options(stringify=True)
{'option1': 'test', 'option2': '2012-01-01'}

:copyright:
    Mazama Science, IRIS
:license:
    GNU Lesser General Public License, Version 3
    (http://www.gnu.org/copyleft/lesser.html)
"""

from obspy.core.utcdatetime import UTCDateTime
from distutils.util import strtobool
from logging import getLogger
from numbers import Number
from datetime import timedelta
from future.utils import with_metaclass

"""
"""

LOGGER = getLogger(__name__)


class Option(object):
    """
    Defines a web service query parameter.  This allows the service API to take parameter
    values as Python objects, and convert them to the proper string form for the service.
    """
    def __init__(self, option_name=None, default=None, hidden=False):
        """
        :param options_name: The name of the option (eg. "starttime")
        :param default: The default to use if no value is set
        :param hidden: Indicate that this option shouldn't be included in the actual query
        """
        self.option_name = option_name
        self.default = default
        self.hidden = hidden

    def to_option(self, value):
        """
        Turn the value into a "native" option type
        """
        return value

    def to_string(self, value):
        return str(value)


class DateOption(Option):
    """
    A date parameter.
    """
    def to_option(self, value):
        # Value can be a number indicating number of days relative to today
        if isinstance(value, Number):
            return UTCDateTime() + timedelta(days=value)
        else:
            return UTCDateTime(value)

    def to_string(self, value):
        if hasattr(value, 'isoformat'):
            return value.isoformat()
        elif isinstance(value, str):
            return value
        else:
            raise ValueError("%s is not a date" % (value,))


class BinaryOption(Option):
    """
    A binary parameter that is passed in a query as "yes" or "no".
    """
    def to_option(self, value):
        return strtobool(value)

    def to_string(self, value):
        return "yes" if value else "no"


class IntOption(Option):
    """
    An integer value
    """
    def to_option(self, value):
        return int(value)


class FloatOption(Option):
    """
    A floating point value
    """
    def to_option(self, value):
        return float(value)


class OptionsMeta(type):
    """
    Metaclass for the Options type, this converts the Options defined in the class attributes
    into type-aware attributes.
    """
    def __new__(cls, name, bases, attrs):
        options = {}
        for attr, option in attrs.items():
            if isinstance(option, Option):
                # Keep the definition
                options[attr] = option
                # Set the actual instance attribute to the default value or None
                if option.default is not None:
                    attrs[attr] = option.to_option(option.default)
                else:
                    attrs[attr] = None
        # Store the option definitions in a private attribute
        attrs['_options'] = options
        return type.__new__(cls, name, bases, attrs)


class Options(with_metaclass(OptionsMeta, object)):
    """
    Base class for a web service request.
    """

    def set_options(self, options):
        """
        Set all the options in the given dictionary
        """
        for k, v in options.items():
            if k in self._options:
                setattr(self, k, v)

    def get_options(self, keys=None, hidden=True, stringify=False):
        """
        Return the options as a dictionary, with a few options
        @param keys: only include the options named
        @param hidden: if True, include options where Option.hidden is set
        @param stringify: convert the option value to a string
        """
        options = {}
        if not keys:
            keys = self.keys(hidden=hidden)
        for attr in keys:
            option = self._options.get(attr)
            if option:
                value = getattr(self, attr, None)
                if value is not None:
                    if stringify:
                        options[attr] = option.to_string(value)
                    else:
                        options[attr] = value
        return options

    def keys(self, hidden=True):
        """
        Return a list of keys

        @param hidden: if True, include options where Option.hidden is set
        """
        return [key for key, option in self._options.items() if hidden or not option.hidden]

    def __setattr__(self, k, v):
        if k in self._options:
            v = self._options[k].to_option(v)
        self.__dict__[k] = v

    def __repr__(self):
        return repr(self.get_options())



# -*- coding: utf-8 -*-
"""
Display logging messages in a Qt textEdit widget

http://stackoverflow.com/questions/28655198/best-way-to-display-logs-in-pyqt

:copyright:
    Mazama Science, IRIS
:license:
    GNU Lesser General Public License, Version 3
    (http://www.gnu.org/copyleft/lesser.html)
"""

import logging


class MyTextEditLoggingHandler(logging.Handler):
    def __init__(self, signal=None):
        super(MyTextEditLoggingHandler, self).__init__(logging.INFO)
        self.signal = signal

    def emit(self, record):
        msg = self.format(record)
        self.signal.emit(msg)



import os.path

__pkg_path___ = os.path.dirname(os.path.abspath(__file__))
# Use Python semantic versioning
# https://packaging.python.org/tutorials/distributing-packages/#semantic-versioning-preferred
__version__ = '1.0.1'
__app_name__ = "PyWEED"



# -*- coding: utf-8 -*-
"""
Common library for adding data to a QTableWidget.

:copyright:
    Mazama Science, IRIS
:license:
    GNU Lesser General Public License, Version 3
    (http://www.gnu.org/copyleft/lesser.html)
"""

from pyweed.gui.MyNumericTableWidgetItem import MyNumericTableWidgetItem
from PyQt4 import QtGui, QtCore
from logging import getLogger

LOGGER = getLogger(__name__)


class Column(object):
    """ Defines a column for `TableItems` """
    def __init__(self, label, description=None, width=None):
        """
        :param label: The label to display in the column header
        :param description: If set, text will appear when the user hovers over the header
        :param width: Hardcode the column width, this is useful for fields that may have long values that
            would otherwise blow out the table width
        """
        self.label = label
        self.description = description
        self.width = width


class TableItems(object):
    """
    Base helper class for adding a bunch of records to a QTableWidget.

    Subclasses just need to define the columns and how to turn a list of
    records into table cells.

    :example:

    >>> class ExampleTableItems(TableItems):
    >>>     columns = [
    >>>         Column('Id'),
    >>>         Column('Name', width=200),
    >>>         Column('Email'),
    >>>         Column('Age'),
    >>>     ]
    >>>     def rows(self, data):
    >>>         for person in data:
    >>>             yield [
    >>>                 self.stringWidget(person.id),
    >>>                 self.stringWidget(person.name),
    >>>                 self.stringWidget(person.email),
    >>>                 self.numericWidget(person.age),
    >>>             ]

    >>> items = ExampleTableItems(q_table_widget)
    >>> items.fill(list_of_people)

    """

    #: Subclass should define the columns
    columns = None
    #: Table that the items are tied to
    table = None
    #: Subclass can set this to enforce a fixed row height (otherwise it will be sized to fit when it gets filled)
    rowHeight = None

    def __init__(self, table, *args):
        self.table = table

    def rows(self, data):
        """
        Turn the data into rows (an iterable of lists) of QTableWidgetItems
        Subclasses should implement this
        """
        pass

    def applyProps(self, widget, **props):
        """
        Apply props to a newly created widget.

        This is used by the various widget methods to handle arbitrary keyword arguments by
        translating them into Qt setter calls.

        For example, passing `textAlignment=QtCore.Qt.AlignCenter` as a keyword argument will result in
        `widget.setTextAlignment(QtCore.Qt.AlignCenter)`
        """
        # Look for Qt setter for any given prop name
        for prop, value in props.items():
            setter = 'set%s%s' % (prop[:1].capitalize(), prop[1:])
            if hasattr(widget, setter):
                try:
                    getattr(widget, setter)(value)
                except Exception as e:
                    LOGGER.error("Tried and failed to set %s: %s", prop, e)
        return widget

    def stringWidget(self, s, **props):
        """ Create a new item displaying the given string """
        return self.applyProps(QtGui.QTableWidgetItem(s), **props)

    def numericWidget(self, i, text=None, **props):
        """
        Create a new item displaying the given numeric value. Pass a string or a string format as `text` to customize
        how the number is actually displayed. The numeric value will still be used for sorting.
        """
        if text is None:
            text = "%s"
        if '%' in text:
            text = text % i
        return self.applyProps(MyNumericTableWidgetItem(i, text), **props)

    def checkboxWidget(self, b, **props):
        """ Create a new checkbox widget showing the given boolean state """
        checkboxItem = self.applyProps(QtGui.QTableWidgetItem(), **props)
        checkboxItem.setFlags(QtCore.Qt.ItemIsEnabled)
        if b:
            checkboxItem.setCheckState(QtCore.Qt.Checked)
        else:
            checkboxItem.setCheckState(QtCore.Qt.Unchecked)
        return checkboxItem

    def initColumns(self):
        # Only run this if the table columns aren't already set up
        if self.table.columnCount() != len(self.columns):
            self.table.setColumnCount(len(self.columns))

            columnLabels = [c.label for c in self.columns]
            self.table.setHorizontalHeaderLabels(columnLabels)

            # Use the first column for identification
            self.table.setColumnHidden(0, True)

            # Set the tooltips
            for i, column in enumerate(self.columns):
                if column.description:
                    self.table.horizontalHeaderItem(i).setToolTip(column.description)

    def fill(self, data):
        """
        Fill the table
        """
        # Clear existing contents
        self.table.setRowCount(0)

        # Initialize the table columns if needed
        self.initColumns()

        # Need to turn off sorting before inserting items
        self.table.setSortingEnabled(False)

        # Add new contents
        for rowidx, row in enumerate(self.rows(data)):
            self.table.insertRow(rowidx)
            if len(row) != len(self.columns):
                LOGGER.error("Row length doesn't match column count: %s / %s", str(row), str([c.label for c in self.columns]))
            for cellidx, cell in enumerate(row):
                self.table.setItem(rowidx, cellidx, cell)

        # Turn sorting back on
        self.table.setSortingEnabled(True)

        # Adjust the row heights
        if self.rowHeight:
            for i in range(self.table.rowCount()):
                self.table.setRowHeight(i, self.rowHeight)
        else:
            self.table.resizeRowsToContents()

        # Adjust the column widths
        for i, column in enumerate(self.columns):
            if column.width:
                self.table.setColumnWidth(i, column.width)
            else:
                self.table.resizeColumnToContents(i)

    def filter(self, filterFn):
        for row in range(self.table.rowCount()):
            if not filterFn or filterFn(row):
                self.table.showRow(row)
            else:
                self.table.hideRow(row)



# -*- coding: utf-8 -*-
"""
Seismap is a subclass of Basemap, specialized for handling seismic data.

Ideas and code borrowed from:
 * https://github.com/matplotlib/basemap/blob/master/examples/allskymap.py

:copyright:
    Mazama Science, IRIS
:license:
    GNU Lesser General Public License, Version 3
    (http://www.gnu.org/copyleft/lesser.html)
"""

from __future__ import (absolute_import, division, print_function)

import numpy as np

from mpl_toolkits.basemap import Basemap
from pyweed.pyweed_utils import get_bounding_circle, get_preferred_origin, get_distance
from logging import getLogger
from PyQt4 import QtCore
from PyQt4.QtCore import pyqtSignal

LOGGER = getLogger(__name__)


class MapMarkers(object):
    """
    Configures and tracks the markers (for items on the map and bounding boxes/toroids) on the map.
    This handles a single type, so there is one instance of this class for events and one for stations.
    """
    color = '#FFFFFF'
    highlight_color = '#FFFF00'
    marker_type = 'o'
    marker_size = 6
    highlight_marker_size = 12
    bounds_linewidth = 2
    bounds_linestyle = 'dashed'
    bounds_alpha = 0.5
    markers = None
    highlights = None
    box_markers = None
    toroid_markers = None

    def __init__(self):
        self.markers = []
        self.highlights = []
        self.box_markers = []
        self.toroid_markers = []


class EventMarkers(MapMarkers):
    color = '#FFFF00'
    highlight_color = '#FFA500'
    marker_type = 'o'  # circle


class StationMarkers(MapMarkers):
    color = '#FF0000'
    highlight_color = '#CD0000'
    marker_type = 'v'  # inverted triangle


class DrawEvent(object):
    """
    Event emitted when drawing on the map starts/stops.
    """
    def __init__(self, mode, points=None):
        #: The drawing mode being started/stopped
        self.mode = mode
        #: When stopping, we may pass back points representing the drawn area
        self.points = points


class Seismap(QtCore.QObject):
    """
    Map display using Basemap
    """
    # Default kwargs
    DEFAULT_BASEMAP_KWARGS = dict(
        projection='cyl',  # We only support cylindrical projection for now
        resolution='l',  # c/l/i/h/f
    )

    # Signals emitted when starting/ending a draw operation
    drawStart = pyqtSignal(object)
    drawEnd = pyqtSignal(object)

    # Cursors for pan and draw modes
    panCursor = QtCore.Qt.PointingHandCursor
    drawCursor = QtCore.Qt.CrossCursor

    # Markers
    eventMarkers = None
    stationMarkers = None

    def __init__(self, canvas):
        super(Seismap, self).__init__()
        self.canvas = canvas
        self.mapFigure = canvas.fig
        self.mapAxes = self.mapFigure.add_axes([0.01, 0.01, .98, .98])

        # Trackers for the markers
        self.eventMarkers = EventMarkers()
        self.stationMarkers = StationMarkers()

        # Basic map features
        self.initBasemap()

        self.initDrawing()

    def initBasemap(self):
        # NOTE:  http://matplotlib.org/basemap/api/basemap_api.html
        # NOTE:  https://gist.github.com/dannguyen/eb1c4e70565d8cb82d63
        self.mapAxes.clear()

        basemap_kwargs = {}
        basemap_kwargs.update(self.DEFAULT_BASEMAP_KWARGS)
        basemap_kwargs.update(
            ax=self.mapAxes
        )
        self.basemap = Basemap(**basemap_kwargs)

        self.basemap.bluemarble(scale=0.25, alpha=0.42)
        self.basemap.drawcoastlines(color='#555566', linewidth=1)
        self.basemap.drawmeridians(np.arange(0, 360, 30))
        self.basemap.drawparallels(np.arange(-90, 90, 30))

        self.canvas.draw_idle()

    def getLatLon(self, x, y):
        """
        Translate a canvas x/y coordinate to lat/lon
        """
        (lon, lat) = self.basemap(x, y, inverse=True)
        return (lat, lon)

    def addMarkers(self, markers, points):
        """
        Display marker locations

        @param markers: either self.eventMarkers or self.stationMarkers
        @param points: a list of (lat, lon) values
        """

        self.clearMarkers(markers, redraw=False)
        self.clearHighlights(markers, redraw=False)

        # See http://matplotlib.org/api/markers_api.html#module-matplotlib.markers
        if len(points):
            (lats, lons) = zip(*points)
            # Plot in projection coordinates
            x, y = self.basemap(lons, lats)
            markers.markers.extend(
                self.basemap.plot(
                    x, y, linestyle='None', marker=markers.marker_type, markersize=markers.marker_size,
                    color=markers.color, markeredgecolor=markers.color
                )
            )

        self.canvas.draw_idle()

    def addHighlights(self, markers, points):
        """
        Highlight selected items

        @param markers: either self.eventMarkers or self.stationMarkers
        @param points: a list of (lat, lon) values
        """

        self.clearHighlights(markers, redraw=False)

        if len(points):
            (lats, lons) = zip(*points)
            # Plot in projection coordinates
            # TODO:  Use self.scatter() with zorder=99 to keep highlighting on top?
            x, y = self.basemap(lons, lats)
            markers.highlights.extend(
                self.basemap.plot(
                    x, y, linestyle='None', marker=markers.marker_type, markersize=markers.highlight_marker_size,
                    color=markers.highlight_color
                )
            )

        self.canvas.draw_idle()

    def addMarkerBox(self, markers, n, e, s, w):
        """
        Display a bounding box
        """
        # Check for box wrapping around the map edge
        if e < w:
            # Need to create two paths
            paths = [
                [[s, 180], [s, w], [n, w], [n, 180]],
                [[s, -180], [s, e], [n, e], [n, -180]]
            ]
        else:
            paths = [
                [[n, w], [n, e], [s, e], [s, w], [n, w]]
            ]

        for path in paths:
            (lats, lons) = zip(*path)
            (x, y) = self.basemap(lons, lats)
            markers.box_markers.extend(
                self.basemap.plot(
                    x, y,
                    color=markers.color,
                    linewidth=markers.bounds_linewidth,
                    linestyle=markers.bounds_linestyle,
                    alpha=markers.bounds_alpha
                ))

        self.canvas.draw_idle()

    def addMarkerToroid(self, markers, lat, lon, minradius, maxradius):
        """
        Display a toroidal bounding area
        """
        for r in (minradius, maxradius):
            if r > 0:
                paths = self.wrapPath(get_bounding_circle(lat, lon, r))
                for path in paths:
                    (lats, lons) = zip(*path)
                    (x, y) = self.basemap(lons, lats)
                    markers.toroid_markers.extend(
                        self.basemap.plot(
                            x, y,
                            color=markers.color,
                            linewidth=markers.bounds_linewidth,
                            linestyle=markers.bounds_linestyle,
                            alpha=markers.bounds_alpha
                        ))

        self.canvas.draw_idle()

    def clearMarkers(self, markers, redraw=True):
        """
        Remove existing marker elements
        """
        # See http://stackoverflow.com/questions/4981815/how-to-remove-lines-in-a-matplotlib-plot
        try:
            while markers.markers:
                markers.markers.pop(0).remove()
            markers.markers = []
        except IndexError:
            pass
        if redraw:
            self.canvas.draw_idle()

    def clearHighlights(self, markers, redraw=True):
        """
        Remove existing highlights
        """
        try:
            while markers.highlights:
                markers.highlights.pop(0).remove()
            markers.highlights = []
        except IndexError:
            pass
        if redraw:
            self.canvas.draw_idle()

    def clearBoundingMarkers(self, markers, redraw=True):
        """
        Remove existing bounding box/toroid markers
        """
        try:
            while markers.box_markers:
                markers.box_markers.pop(0).remove()
            markers.box_markers = []
            while markers.toroid_markers:
                markers.toroid_markers.pop(0).remove()
            markers.toroid_markers = []
        except IndexError:
            pass
        if redraw:
            self.canvas.draw_idle()

    def addEvents(self, catalog):
        """
        Display event locations
        """
        points = [
            (o.latitude, o.longitude)
            for o in
            [get_preferred_origin(e) for e in catalog]
            if o
        ]
        self.addMarkers(self.eventMarkers, points)

    def addEventsHighlighting(self, points):
        """
        Highlights selected events
        """
        self.addHighlights(self.eventMarkers, points)

    def addEventsBox(self, n, e, s, w):
        """
        Display event box
        """
        self.addMarkerBox(self.eventMarkers, n, e, s, w)

    def addEventsToroid(self, lat, lon, minradius, maxradius):
        """
        Display event locations
        """
        self.addMarkerToroid(self.eventMarkers, lat, lon, minradius, maxradius)

    def clearEventsBounds(self):
        """
        Clear event bounding markers
        """
        self.clearBoundingMarkers(self.eventMarkers)

    def addStations(self, inventory):
        """
        Display station locations
        """
        points = [
            (s.latitude, s.longitude)
            for n in inventory.networks
            for s in n.stations
        ]
        self.addMarkers(self.stationMarkers, points)

    def addStationsHighlighting(self, points):
        """
        Highlight selected stations
        """
        self.addHighlights(self.stationMarkers, points)

    def addStationsBox(self, n, e, s, w):
        """
        Display station box outline
        """
        self.addMarkerBox(self.stationMarkers, n, e, s, w)

    def addStationsToroid(self, lat, lon, minradius, maxradius):
        """
        Display station locations
        """
        self.addMarkerToroid(self.stationMarkers, lat, lon, minradius, maxradius)

    def clearStationsBounds(self):
        """
        Clear station bounding markers
        """
        self.clearBoundingMarkers(self.stationMarkers)

    def wrapPath(self, path):
        """
        Split a path of (lat, lon) values into a list of paths none of which crosses the map boundary

        This is a workaround for https://github.com/matplotlib/basemap/issues/214
        """
        lastPoint = None
        currentSubpath = []
        subpaths = [currentSubpath]
        for point in path:
            # Detect the path wrapping around the map boundary
            midpoints = self.wrapPoint(lastPoint, point)
            if midpoints:
                # The first midpoint goes at the end of the previous subpath
                currentSubpath.append(midpoints[0])
                currentSubpath = []
                subpaths.append(currentSubpath)
                # The second midpoint goes at the start of the new subpath
                currentSubpath.append(midpoints[1])
            currentSubpath.append(point)
            lastPoint = point
        # If the path crosses the map boundary, we will (usually?) have 3 subpaths, and the
        # first and last subpaths are actually contiguous so join them together
        if len(subpaths) > 2:
            subpaths[-1].extend(subpaths[0])
            subpaths = subpaths[1:]
        return subpaths

    def wrapPoint(self, p1, p2):
        """
        Given two lat/lon pairs representing a line segment to be plotted, if the segment crosses
        the map boundary this returns two points (one at each edge of the map) where the segment
        crosses the boundary, or None if the segment doesn't cross the boundary.

        For example:
        wrapPoint([20, 170], [10, -170])
        returns
        [[15, 180], [15, -180]]

        This means that the segment should be split into:
        [20, 170] - [15, 180]
        [15, -180] - [10, -170]
        """
        if p1 and p2:
            dlon = p2[1] - p1[1]
            if abs(dlon) > 180:
                # Interpolate the point where the segment crosses the boundary
                dlat = p2[0] - p1[0]
                if dlon < 0:
                    # Wraps from east to west
                    midlat = p1[0] + (180 - p1[1]) * dlat / (dlon + 360)
                    return [[midlat, 180], [midlat, -180]]
                else:
                    # Wraps from west to east
                    midlat = p1[0] + (-180 - p1[1]) * dlat / (dlon - 360)
                    return [[midlat, -180], [midlat, 180]]
        return None

    def zoomIn(self):
        """
        Zoom into the map
        """
        # Zoom in by 2, by trimming 1/4 from each edge
        xlim = self.mapAxes.get_xlim()
        xdelta = (xlim[1] - xlim[0]) / 4
        ylim = self.mapAxes.get_ylim()
        ydelta = (ylim[1] - ylim[0]) / 4

        self.mapAxes.set_xlim(xlim[0] + xdelta, xlim[1] - xdelta)
        self.mapAxes.set_ylim(ylim[0] + ydelta, ylim[1] - ydelta)
        self.canvas.draw_idle()

    def zoomOut(self):
        """
        Zoom out on the map
        """
        # Zoom out by 2, by adding 1/2 to each edge
        xlim = self.mapAxes.get_xlim()
        xdelta = (xlim[1] - xlim[0]) / 2
        ylim = self.mapAxes.get_ylim()
        ydelta = (ylim[1] - ylim[0]) / 2

        self.mapAxes.set_xlim(xlim[0] - xdelta, xlim[1] + xdelta)
        self.mapAxes.set_ylim(ylim[0] - ydelta, ylim[1] + ydelta)
        self.canvas.draw_idle()

    def zoomReset(self):
        """
        Zoom all the way out of the map
        """
        self.mapAxes.set_xlim(-180, 180)
        self.mapAxes.set_ylim(-90, 90)
        self.canvas.draw_idle()

    def fitCanvas(self, *args):
        canvas_w = self.canvas.width()
        canvas_h = self.canvas.height()
        map_xlim = self.mapAxes.get_xlim()
        map_ylim = self.mapAxes.get_ylim()

        map_w = map_xlim[1] - map_xlim[0]
        map_h = map_ylim[1] - map_ylim[0]
        prop_map_h = ((map_w * canvas_h) / canvas_w)

        adjustment = ((prop_map_h - map_h) / 2)

        self.mapAxes.set_ylim(map_ylim[0] - adjustment, map_ylim[1] + adjustment)
        self.canvas.draw_idle()

    def initDrawing(self):
        # Map drawing mode
        self.draw_mode = None
        # Indicates that we are actually drawing (ie. mouse button is down)
        self.drawing = False
        # If drawing, the start and end points
        self.draw_points = []
        # Handler functions for mouse down/move/up, these are created when drawing is activated
        # See http://matplotlib.org/users/event_handling.html
        self.draw_handlers = {
            'click': self.canvas.mpl_connect('button_press_event', self.onMouseDown),
            'scroll_wheel': self.canvas.mpl_connect('scroll_event', self.onScrollWheel),
            'resize': self.canvas.mpl_connect('resize_event', self.fitCanvas),
        }
        self.updateCursor()

    def setDrawMode(self, mode):
        """
        Initialize the given drawing mode
        """
        LOGGER.info("Drawing mode set to %s", mode)
        self.drawing = False
        self.draw_mode = mode
        self.drawStart.emit(DrawEvent(mode))
        self.updateCursor()

    def clearDrawMode(self):
        """
        Clear any active drawing mode
        """
        if self.draw_mode:
            points = None
            if self.drawing:
                # If we finished drawing bounds, pass back the parameters indicated
                if self.draw_points:
                    if 'box' in self.draw_mode:
                        points = self.drawPointsToBox()
                    elif 'toroid' in self.draw_mode:
                        points = self.drawPointsToToroid()
            self.drawEnd.emit(DrawEvent(self.draw_mode, points))
        self.drawing = False
        self.draw_mode = None
        self.updateCursor()

    def toggleDrawMode(self, mode, toggle):
        """
        Event handler when a draw mode button is clicked
        """
        if toggle:
            self.setDrawMode(mode)
        else:
            self.clearDrawMode()

    def updateCursor(self):
        """
        Set the cursor on the canvas according to the current mode
        """
        if self.draw_mode:
            self.canvas.setCursor(self.drawCursor)
        else:
            self.canvas.setCursor(self.panCursor)

    def drawPointsToBox(self):
        """
        Convert self.draw_points to a tuple of (n, e, s, w)
        """
        (lat1, lon1) = self.draw_points[0]
        (lat2, lon2) = self.draw_points[1]
        return (
            max(lat1, lat2),
            max(lon1, lon2),
            min(lat1, lat2),
            min(lon1, lon2)
        )

    def drawPointsToToroid(self):
        """
        Convert self.draw_points to a tuple of (lat, lon, radius)
        """
        (lat1, lon1) = self.draw_points[0]
        (lat2, lon2) = self.draw_points[1]
        radius = get_distance(lat1, lon1, lat2, lon2)
        return (lat1, lon1, radius)

    def onMouseDown(self, event):
        """
        Handle a mouse click on the map
        """
        if self.draw_mode:
            # Start a drawing operation
            (lat, lon) = self.getLatLon(event.xdata, event.ydata)
            if lat is not None and lon is not None:
                self.drawing = True
                self.draw_points = [[lat, lon], [lat, lon]]
        else:
            # If there's no draw operation, pan the map
            if event.inaxes == self.mapAxes:
                LOGGER.debug("Starting pan")
                self.drawing = True
                self.mapAxes.start_pan(event.x, event.y, event.button)
        # If we've started a drawing operation, connect listeners for activity
        if self.drawing:
            self.draw_handlers['release'] = self.canvas.mpl_connect('button_release_event', self.onMouseUp)
            self.draw_handlers['move'] = self.canvas.mpl_connect('motion_notify_event', self.onMouseMove)

    def onMouseUp(self, event):
        """
        Handle a mouse up event, this should only be called while the user is drawing on the map
        """
        # Disconnect the release/move handlers
        for event in ('release', 'move',):
            if event in self.draw_handlers:
                self.canvas.mpl_disconnect(self.draw_handlers[event])
                del self.draw_handlers[event]

        # If there's no draw mode, we are probably panning so turn that off
        if not self.draw_mode:
            self.mapAxes.end_pan()

        # Exit drawing mode
        self.clearDrawMode()

    def onMouseMove(self, event):
        """
        Handle a mouse move event, this should only be called while the user is drawing on the map
        """
        if self.drawing:
            if self.draw_mode:
                (lat, lon) = self.getLatLon(event.xdata, event.ydata)
                if lat is not None and lon is not None:
                    self.draw_points[1] = [lat, lon]
                    LOGGER.debug("Draw points: %s" % self.draw_points)
                    self.updateDrawBounds()
            else:
                # If we aren't drawing anything, pan the map
                self.mapAxes.drag_pan(event.button, event.key, event.x, event.y)
                self.canvas.draw_idle()

    def updateDrawBounds(self):
        """
        Update the displayed bounding box/toroid as the user is drawing it
        """
        # Clear any existing bounds
        if 'events' in self.draw_mode:
            self.clearEventsBounds()
        elif 'stations' in self.draw_mode:
            self.clearStationsBounds()
        # Build options values based on box or toroid
        if 'box' in self.draw_mode:
            (n, e, s, w) = self.drawPointsToBox()
            if 'events' in self.draw_mode:
                self.addEventsBox(n, e, s, w)
            elif 'stations' in self.draw_mode:
                self.addStationsBox(n, e, s, w)
        elif 'toroid' in self.draw_mode:
            (lat, lon, maxradius) = self.drawPointsToToroid()
            if 'events' in self.draw_mode:
                self.addEventsToroid(lat, lon, 0, maxradius)
            elif 'stations' in self.draw_mode:
                self.addStationsToroid(lat, lon, 0, maxradius)

    def onScrollWheel(self, event):
        """
        Zoom on scroll wheel
        """
        if event.step:
            if event.step > 0:
                self.zoomIn()
            else:
                self.zoomOut()



# -*- coding: utf-8 -*-
"""
Station Options

:copyright:
    Mazama Science, IRIS
:license:
    GNU Lesser General Public License, Version 3
    (http://www.gnu.org/copyleft/lesser.html)
"""

from PyQt4 import QtGui, QtCore
from pyweed.gui.uic import StationOptionsWidget
import logging
from pyweed.station_options import StationOptions
from distutils.util import strtobool
from pyweed.gui.OptionsWidget import BaseOptionsWidget

LOGGER = logging.getLogger(__name__)


class StationOptionsWidget(BaseOptionsWidget, StationOptionsWidget.Ui_StationOptionsWidget):
    """
    Dialog window for event options used in creating a webservice query.
    """

    def mapInputs(self):
        return {
            'network': self.networkLineEdit,
            'station': self.stationLineEdit,
            'location': self.locationLineEdit,
            'channel': self.channelLineEdit,
            'starttime': self.starttimeDateTimeEdit,
            'endtime': self.endtimeDateTimeEdit,
            'minlongitude': self.locationRangeWestDoubleSpinBox,
            'maxlongitude': self.locationRangeEastDoubleSpinBox,
            'minlatitude': self.locationRangeSouthDoubleSpinBox,
            'maxlatitude': self.locationRangeNorthDoubleSpinBox,
            'minradius': self.distanceFromPointMinRadiusDoubleSpinBox,
            'maxradius': self.distanceFromPointMaxRadiusDoubleSpinBox,
            'longitude': self.distanceFromPointEastDoubleSpinBox,
            'latitude': self.distanceFromPointNorthDoubleSpinBox,
            '_locationGlobal': self.locationGlobalRadioButton,
            '_locationRange': self.locationRangeRadioButton,
            '_locationDistanceFromPoint': self.locationDistanceFromPointRadioButton,
            '_locationDistanceFromEvents': self.locationDistanceFromEventsRadioButton,
            'mindistance': self.distanceFromEventsMinDoubleSpinBox,
            'maxdistance': self.distanceFromEventsMaxDoubleSpinBox,
        }

    def optionsToInputs(self, values):
        # Turn the option choice value into a set of radio values
        location_choice = values.get('location_choice')
        values['_locationGlobal'] = (location_choice == StationOptions.LOCATION_GLOBAL)
        values['_locationRange'] = (location_choice == StationOptions.LOCATION_BOX)
        values['_locationDistanceFromPoint'] = (location_choice == StationOptions.LOCATION_POINT)
        values['_locationDistanceFromEvents'] = (location_choice == StationOptions.LOCATION_EVENTS)
        return values

    def inputsToOptions(self, values):
        # Turn the radio values into an option choice value
        if strtobool(values.get('_locationGlobal', 'F')):
            values['location_choice'] = StationOptions.LOCATION_GLOBAL
        elif strtobool(values.get('_locationRange', 'F')):
            values['location_choice'] = StationOptions.LOCATION_BOX
        elif strtobool(values.get('_locationDistanceFromPoint', 'F')):
            values['location_choice'] = StationOptions.LOCATION_POINT
        elif strtobool(values.get('_locationDistanceFromEvents', 'F')):
            values['location_choice'] = StationOptions.LOCATION_EVENTS
        return values

    def get_timeFromOtherButton(self):
        return self.timeFromEventsToolButton

    def get_locationFromOtherButton(self):
        return self.locationFromEventsToolButton

    def onEventSelectionChanged(self):
        """
        This should be called whenever the event selection has changed.
        If the "distance from selected events" is enabled, this will emit a
        change event.
        """
        key = '_locationDistanceFromEvents'
        inputValue = self.getInputValue(key)
        LOGGER.debug("StationOptionsWidget.onEventSelectionChanged: %s", inputValue)
        if inputValue and strtobool(inputValue):
            self.changed.emit(key)
            self.changedCoords.emit(key)



# -*- coding: utf-8 -*-
"""
Custom widget that embeds an image

To be inserted into the QTableWidget with setCellWidget() rather than setItem().

http://stackoverflow.com/questions/5553342/adding-images-to-a-qtablewidget-in-pyqt

:copyright:
    Mazama Science, IRIS
:license:
    GNU Lesser General Public License, Version 3
    (http://www.gnu.org/copyleft/lesser.html)
"""

from PyQt4 import QtGui


class MyTableWidgetImageWidget(QtGui.QLabel):

    def __init__(self, parent=None, imagePath=None):
        super(MyTableWidgetImageWidget, self).__init__(parent)
        pic = QtGui.QPixmap(imagePath)
        # NOTE:  From QPixmap we could use scaledToWidth() but the resulting images aren't so pretty
        # pic = pic.scaledToWidth(500)
        self.setPixmap(pic)
        # NOTE:  From QLabel we could use setScaledContents() but this doesn't preserve the aspect ratio
        # self.setScaledContents(True)



# -*- coding: utf-8 -*-
"""
Spinner overlay to indicate that an operation is in progress.

:copyright:
    Mazama Science, IRIS
:license:
    GNU Lesser General Public License, Version 3
    (http://www.gnu.org/copyleft/lesser.html)
"""

from PyQt4 import QtGui, QtCore
from logging import getLogger
from pyweed.gui.uic import SpinnerWidget

LOGGER = getLogger(__name__)


class SpinnerWidget(QtGui.QFrame, SpinnerWidget.Ui_SpinnerWidget):
    """
    Spinner overlay widget. When shown, this covers the parent window and
    plays an animated spinner graphic along with a text message.
    """

    cancelled = QtCore.pyqtSignal()

    def __init__(self, labelText, cancellable=True, parent=None):
        super(SpinnerWidget, self).__init__(parent=parent)
        self.setupUi(self)
        self.movie = QtGui.QMovie(":qrc/rotator-32.gif")
        self.icon.setMovie(self.movie)
        self.originalLabelText = labelText
        self.label.setText(labelText)
        if cancellable:
            self.cancelButton.clicked.connect(self.cancelled.emit)
        else:
            self.cancelButton.hide()
        self.hide()

    def setLabel(self, labelText):
        self.label.setText(labelText)

    def showEvent(self, *args, **kwargs):
        """
        Widget is being shown
        """
        self.setGeometry(self.parent().contentsRect())
        self.movie.start()
        self.label.setText(self.originalLabelText)
        return super(SpinnerWidget, self).showEvent(*args, **kwargs)

    def hideEvent(self, *args, **kwargs):
        """
        Widget is being hidden
        """
        self.movie.stop()
        return super(SpinnerWidget, self).hideEvent(*args, **kwargs)



# -*- coding: utf-8 -*-
"""
Custom QTableWidgetItem that forces numerical sorting

http://stackoverflow.com/questions/25533140/sorting-qtablewidget-items-numerically

:copyright:
    Mazama Science, IRIS
:license:
    GNU Lesser General Public License, Version 3
    (http://www.gnu.org/copyleft/lesser.html)
"""

from PyQt4 import QtCore
from PyQt4 import QtGui
from pyweed.gui.TableWidget import CustomTableWidgetItemMixin


class MyNumericTableWidgetItem (CustomTableWidgetItemMixin, QtGui.QTableWidgetItem):
    """
    Custom QTableWidgetItem that forces numerical sorting

    http://stackoverflow.com/questions/25533140/sorting-qtablewidget-items-numerically
    """

    def __init__(self, value, text):
        super(MyNumericTableWidgetItem, self).__init__(text)
        self.setData(QtCore.Qt.UserRole, value)

    def __lt__(self, other):
        if (isinstance(other, MyNumericTableWidgetItem)):
            selfDataValue = float(self.data(QtCore.Qt.UserRole))
            otherDataValue = float(other.data(QtCore.Qt.UserRole))
            return selfDataValue < otherDataValue
        else:
            return QtGui.QTableWidgetItem.__lt__(self, other)



# -*- coding: utf-8 -*-
"""
Container for waveforms.

:copyright:
    Mazama Science, IRIS
:license:
    GNU Lesser General Public License, Version 3
    (http://www.gnu.org/copyleft/lesser.html)
"""

from __future__ import (absolute_import, division, print_function)

import os
from obspy import UTCDateTime
from pyweed.signals import SignalingThread, SignalingObject
import collections
from PyQt4 import QtCore, QtGui
import obspy
from logging import getLogger
import matplotlib
import weakref
from pyweed.pyweed_utils import get_sncl, get_event_id, TimeWindow,\
    get_preferred_origin, get_preferred_magnitude, OUTPUT_FORMAT_EXTENSIONS,\
    get_event_description, get_arrivals, format_time_str, get_distance, get_service_url, CancelledException
from obspy.core.util.attribdict import AttribDict
from obspy.io.sac.sactrace import SACTrace
from obspy.core.stream import Stream
import concurrent.futures

LOGGER = getLogger(__name__)

THREAD_POOL_SIZE = 10


class WaveformEntry(AttribDict):
    """
    Class representing an event/channel combination and the relevant waveform request
    """
    defaults = dict(
        # Weak refs to ObsPy event/channel
        event_ref=None,
        network_ref=None,
        station_ref=None,
        channel_ref=None,
        # Distance from event to station
        distance=None,
        # Phase arrivals
        arrivals=None,
        # Timing information
        event_time=None,
        start_time=None,
        end_time=None,
        start_string=None,
        end_string=None,
        # Reflects user checkbox in the table
        keep=True,
        # Tracking IDs
        waveform_id=None,
        # Other table display values
        sncl=None,
        event_time_str=None,
        event_description=None,
        event_mag=None,
        event_mag_value=None,
        event_depth=None,
        # Base directory to download files to (from WaveformsHandler)
        download_dir=None,
        # Time window settings (from WaveformHandler)
        time_window=None,
        # Base filename for mseed, image, etc.
        base_filename=None,
        # Full paths
        mseed_path=None,
        mseed_exists=False,
        image_path=None,
        image_exists=False,
        # Loading indicator
        loading=False,
        # Error message if unable to load
        error=None,
    )

    def __init__(self, event, network, station, channel, *args, **kwargs):
        super(WaveformEntry, self).__init__(*args, **kwargs)

        self.event_ref = weakref.ref(event)
        self.network_ref = weakref.ref(network)
        self.station_ref = weakref.ref(station)
        self.channel_ref = weakref.ref(channel)

        self.sncl = get_sncl(network, station, channel)

        self.event_description = get_event_description(event)
        origin = get_preferred_origin(event)
        self.event_time = origin.time
        self.event_time_str = format_time_str(origin.time)
        self.event_depth = origin.depth / 1000
        mag = get_preferred_magnitude(event)
        self.event_mag = "%s%s" % (mag.mag, mag.magnitude_type)
        self.event_mag_value = mag.mag
        self.waveform_id = '%s_%s' % (self.sncl, get_event_id(event))

        self.distance = get_distance(
            origin.latitude, origin.longitude,
            station.latitude, station.longitude)

    def update_handler_values(self, waveform_handler):
        """
        Update any values that come from the WaveformHander
        """
        self.download_dir = waveform_handler.downloadDir
        self.time_window = waveform_handler.time_window

    def prepare(self):
        """
        Calculate (or recalculate) values in preparation for doing work
        """
        if not self.arrivals:
            self.arrivals = get_arrivals(self.distance, self.event_depth)

        (self.start_time, self.end_time) = self.time_window.calculate_window(
            self.event_time, self.arrivals)
        self.start_string = UTCDateTime(self.start_time).format_iris_web_service().replace(':', '_')
        self.end_string = UTCDateTime(self.end_time).format_iris_web_service().replace(':', '_')

        self.base_filename = "%s_%s_%s" % (self.sncl, self.start_string, self.end_string)
        self.mseed_path = os.path.join(self.download_dir, "%s.mseed" % self.base_filename)
        self.image_path = os.path.join(self.download_dir, "%s.png" % self.base_filename)

        self.check_files()

    def check_files(self):
        """
        After calculating the paths for downloaded files, check to see if they're already there
        """
        self.mseed_exists = os.path.exists(self.mseed_path)
        self.image_exists = os.path.exists(self.image_path)


class WaveformResult(object):
    """
    Container for a waveform result to be passed as a signal, includes the waveform ID so that
    the result can be correctly handled.
    """
    def __init__(self, waveform_id, result):
        self.waveform_id = waveform_id
        self.result = result


def load_waveform(client, waveform):
    """
    Download the given waveform data and generate an image. This is a standalone function so we can
    run it in a separate thread. This modifies the waveform entry and returns a dummy value, or
    raises an exception on any error.
    """
    plot_width = 600
    plot_height = 120

    waveform_id = waveform.waveform_id
    LOGGER.debug("Loading waveform: %s (%s)", waveform_id, QtCore.QThread.currentThreadId())
#     LOGGER.debug("Adding slowness")
#     QtCore.QThread.currentThread().sleep(2)

    try:
        waveform.prepare()

        if waveform.image_exists:
            # No download needed
            LOGGER.info("Waveform %s already has an image" % waveform_id)
            return True

        mseedFile = waveform.mseed_path
        LOGGER.debug("%s save as MiniSEED", waveform_id)

        # Load data from disk or network as appropriate
        if os.path.exists(mseedFile):
            LOGGER.info("Loading waveform data for %s from %s", waveform_id, mseedFile)
            st = obspy.read(mseedFile)
        else:
            (network, station, location, channel) = waveform.sncl.split('.')
            service_url = get_service_url(client, 'dataselect', {
                "network": network, "station": station, "location": location, "channel": channel,
                "starttime": waveform.start_time, "endtime": waveform.end_time,
            })
            LOGGER.info("Retrieving waveform data for %s from %s", waveform_id, service_url)
            st = client.get_waveforms(
                network, station, location, channel, waveform.start_string, waveform.end_string)
            # Write to file
            st.write(mseedFile, format="MSEED")

        # Generate image if necessary
        imageFile = waveform.image_path
        if not os.path.exists(imageFile):
            LOGGER.debug('Plotting waveform image to %s', imageFile)
            # In order to really customize the plotting, we need to return the figure and modify it
            h = st.plot(size=(plot_width, plot_height), handle=True)
            # Resize the subplot to a hard size, because otherwise it will do it inconsistently
            h.subplots_adjust(bottom=.2, left=.1, right=.95, top=.95)
            # Remove the title
            for c in h.get_children():
                if isinstance(c, matplotlib.text.Text):
                    c.remove()
            # Save with transparency
            h.savefig(imageFile)
            matplotlib.pyplot.close(h)

        waveform.check_files()

    except Exception as e:
        # Most common error is "no data" TODO: see https://github.com/obspy/obspy/issues/1656
        if str(e).startswith("No data"):
            waveform.error = "No data available"
        else:
            waveform.error = str(e)
        # Reraise the exception to signal an error to the caller
        raise

    finally:
        # Mark as finished loading
        waveform.loading = False

    return True


class WaveformsLoader(SignalingThread):
    """
    Thread to download waveform data and generate an image
    """
    progress = QtCore.pyqtSignal(object)

    def __init__(self, client, waveforms):
        """
        Initialization.
        """
        # Keep a reference to globally shared components
        self.client = client
        self.waveforms = waveforms
        self.futures = {}
        super(WaveformsLoader, self).__init__()

    def run(self):
        """
        Make a webservice request for events using the passed in options.
        """
        self.setPriority(QtCore.QThread.LowestPriority)
        self.clearFutures()
        self.futures = {}
        with concurrent.futures.ThreadPoolExecutor(THREAD_POOL_SIZE) as executor:
            for waveform in self.waveforms:
                # Dictionary to look up the waveform id by Future
                self.futures[executor.submit(load_waveform, self.client, waveform)] = waveform.waveform_id
            # Iterate through Futures as they complete
            for result in concurrent.futures.as_completed(self.futures):
                waveform_id = self.futures.get(result)
                if waveform_id:
                    LOGGER.debug("Loader finished: %s", waveform_id)
                    try:
                        self.progress.emit(WaveformResult(waveform_id, result.result()))
                    except Exception as e:
                        self.progress.emit(WaveformResult(waveform_id, e))
        self.futures = {}
        self.done.emit(None)

    def clearFutures(self):
        """
        Cancel any outstanding tasks
        """
        if self.futures:
            for future in self.futures:
                if not future.done():
                    LOGGER.debug("Cancelling unexecuted future")
                    future.cancel()

    def cancel(self):
        """
        User-requested cancel
        """
        self.done.disconnect()
        self.progress.disconnect()
        self.clearFutures()


class WaveformsHandler(SignalingObject):
    """
    Manage the waveforms retrieval.

    The contents of self.waveforms is determined by the user selections
    in the main GUI events and SNCL tables. There will be a separate
    entry in self.waveforms for each event-SNCL combination regardless
    of any filtering that is applied in the WaveformsDialog to reduce
    the size of the visible table.
    """

    progress = QtCore.pyqtSignal(object)

    def __init__(self, logger, preferences, client):
        """
        Initialization.
        """
        super(WaveformsHandler, self).__init__()

        # Keep a reference to globally shared components
        self.preferences = preferences
        self.client = client

        # Important preferences
        self.downloadDir = self.preferences.Waveforms.downloadDir

        # Loader component
        self.waveforms_loader = None
        # Thread to run the loader in
        self.thread = None
        # Asynchronous requests, each is working to download a single waveform
        self.requests = None

        # A TimeWindow object giving offsets and phase arrivals
        self.time_window = TimeWindow()

        # Current list of waveform entries
        self.waveforms = None
        # Waveforms indexed by id
        self.waveforms_by_id = None

    def create_waveforms(self, pyweed):
        """
        Create a list of waveform entries based on the current event/station selections
        """
        self.waveforms = [
            WaveformEntry(
                event, network, station, channel
            )
            for (event, network, station, channel) in pyweed.iter_selected_events_stations()
        ]
        self.waveforms_by_id = dict(
            (waveform.waveform_id, waveform)
            for waveform in self.waveforms
        )

    def cancel_download(self):
        """
        Cancel the loader if it's running
        """
        LOGGER.debug('Cancelling existing downloads')
        if self.waveforms_loader:
            self.waveforms_loader.cancel()
        # Mark all the waveforms that are still marked as loading
        for waveform in self.waveforms:
            if waveform.loading:
                waveform.loading = False
                waveform.error = "Cancelled download"
                LOGGER.debug("Manually marking %s as cancelled", waveform.waveform_id)
                self.progress.emit(WaveformResult(waveform.waveform_id, None))
        self.done.emit(CancelledException())

    def download_waveforms(self, priority_ids, other_ids, time_window):
        """
        Initiate a download of all the given waveforms
        """
        LOGGER.info('Downloading waveforms')
        LOGGER.debug("Priority IDs: %s" % (priority_ids,))
        LOGGER.debug("Other IDs: %s" % (other_ids,))

        # Prepare the waveform entries
        self.time_window = time_window
        for waveform in self.waveforms:
            waveform.update_handler_values(self)
            # Clear error flag and set loading flag
            waveform.error = None
            waveform.loading = True

        # Get the waveforms ordered by priority
        waveform_ids = list(priority_ids) + list(other_ids)
        waveforms = [self.get_waveform(waveform_id) for waveform_id in waveform_ids]

        # Create a worker to load the data in a separate thread
        self.waveforms_loader = WaveformsLoader(self.client, waveforms)
        self.waveforms_loader.progress.connect(self.on_downloaded)
        self.waveforms_loader.done.connect(self.on_all_downloaded)
        self.waveforms_loader.start()

    def on_downloaded(self, result):
        """
        Called for each downloaded waveform.

        :param result: a `WaveformResult`
        """
        # LOGGER.debug("Downloaded waveform %s (%s)", result.waveform_id, QtCore.QThread.currentThreadId())
        self.progress.emit(result)

    def on_all_downloaded(self, result):
        LOGGER.debug("All waveforms downloaded  (%s)", QtCore.QThread.currentThreadId())
        if self.waveforms_loader:
            self.waveforms_loader.quit()
            self.waveforms_loader.wait()
            LOGGER.debug("Download thread exited")
        self.done.emit(result)

    def get_waveform(self, waveform_id):
        """
        Retrieve the Series for the given waveform
        """
        return self.waveforms_by_id.get(waveform_id)

    def save_waveforms_iter(self, base_output_path, output_format, waveforms):
        """
        Save waveforms, returning an iterator of WaveformResult objects whose values are
        True (saved), False (already saved), or Exception (error)
        This is so the GUI layer can handle save progress and errors in its own fashion
        """
        if not os.path.exists(base_output_path):
            try:
                os.makedirs(base_output_path, 0o700)
            except Exception as e:
                raise Exception("Could not create the output path: %s" % str(e))

        # Get the file extension to use
        extension = OUTPUT_FORMAT_EXTENSIONS[output_format]

        for waveform in waveforms:
            waveform_id = waveform.waveform_id
            try:
                output_file = os.path.extsep.join((waveform.base_filename, extension))
                output_path = os.path.join(base_output_path, output_file)
                # Don't repeat any work that has already been done
                if not os.path.exists(output_path):
                    LOGGER.debug('reading %s', waveform.mseed_path)
                    st = obspy.read(waveform.mseed_path)
                    self.save_waveform(st, output_path, output_format, waveform)
                    yield WaveformResult(waveform_id, True)
                else:
                    yield WaveformResult(waveform_id, False)
            except Exception as e:
                yield WaveformResult(waveform_id, e)

    def save_waveform(self, st, output_path, output_format, waveform):
        """
        Save a single waveform. This should handle any special output features (eg. SAC metadata)
        """
        LOGGER.debug('writing %s', output_path)
        if output_format in ('SAC', 'SACXY',):
            # For SAC output, we need to pull header data from the waveform record
            st = self.to_sac_stream(st, waveform)
        st.write(output_path, format=output_format)

    def to_sac_stream(self, st, waveform):
        """
        Return a SACTrace based on the given stream, containing metadata headers from the waveform
        """
        tr = SACTrace.from_obspy_trace(st[0])
        tr.kevnm = waveform.event_description[:16]
        event = waveform.event_ref()
        if not event:
            LOGGER.warn("Lost reference to event %s", waveform.event_description)
        else:
            origin = get_preferred_origin(waveform.event_ref())
            if origin:
                tr.evla = origin.latitude
                tr.evlo = origin.longitude
                tr.evdp = origin.depth / 1000
                tr.o = origin.time - waveform.start_time
            magnitude = get_preferred_magnitude(waveform.event_ref())
            if magnitude:
                tr.mag = magnitude.mag
        channel = waveform.channel_ref()
        if not channel:
            LOGGER.warn("Lost reference to channel %s", waveform.sncl)
        else:
            tr.stla = channel.latitude
            tr.stlo = channel.longitude
            tr.stdp = channel.depth
            tr.stel = channel.elevation
            tr.cmpaz = channel.azimuth
            tr.cmpinc = channel.dip + 90
            tr.kinst = channel.sensor.description[:8]
        return Stream([tr.to_obspy_trace()])


# ------------------------------------------------------------------------------
# Helper functions
# ------------------------------------------------------------------------------


# ------------------------------------------------------------------------------
# Main
# ------------------------------------------------------------------------------

if __name__ == '__main__':
    import doctest
    doctest.testmod(exclude_empty=True)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
PyWEED GUI

:copyright:
    Mazama Science, IRIS
:license:
    GNU Lesser General Public License, Version 3
    (http://www.gnu.org/copyleft/lesser.html)
"""

from __future__ import (absolute_import, division, print_function)

# Basic packages
import sys
import logging
import os.path
import platform
from PyQt4 import QtCore
from PyQt4 import QtGui

from obspy.core.event.catalog import read_events
from obspy.core.inventory.inventory import read_inventory

from pyweed import __version__, __app_name__
from pyweed.gui.MainWindow import MainWindow
from pyweed.gui.LoggingDialog import LoggingDialog
from pyweed.events_handler import EventsHandler
from pyweed.stations_handler import StationsHandler
from pyweed.gui.WaveformDialog import WaveformDialog
from pyweed.gui.ConsoleDialog import ConsoleDialog
from pyweed.gui.PreferencesDialog import PreferencesDialog
from pyweed.pyweed_core import PyWeedCore
from pyweed.summary import get_filtered_catalog, get_filtered_inventory

LOGGER = logging.getLogger(__name__)


class PyWeedGUI(PyWeedCore, QtCore.QObject):

    # We need to define mainWindow since we may check it before we are fully initialized
    mainWindow = None

    def __init__(self):
        super(PyWeedGUI, self).__init__()

        LOGGER.info('Setting up main window...')
        self.mainWindow = MainWindow(self)

        # Logging
        # see:  http://stackoverflow.com/questions/28655198/best-way-to-display-logs-in-pyqt
        # see:  http://stackoverflow.com/questions/24469662/how-to-redirect-logger-output-into-pyqt-text-widget
        self.loggingDialog = LoggingDialog(self.mainWindow)

        # Waveforms
        # NOTE:  The WaveformsHandler is created inside waveformsDialog.  It is only relevant to that Dialog.
        LOGGER.info('Setting up waveforms dialog...')
        self.waveformsDialog = WaveformDialog(self, self.mainWindow)

        # Preferences
        self.preferencesDialog = PreferencesDialog(self, self.mainWindow)

        # Python console
        self.console = ConsoleDialog(self, self.mainWindow)

        self.configureMenu()

        # Display MainWindow
        LOGGER.info('Showing main window...')
        self.mainWindow.initialize()
        self.mainWindow.show()

    ###############
    # Events
    ###############

    def set_event_options(self, options):
        super(PyWeedGUI, self).set_event_options(options)
        if self.mainWindow:
            self.mainWindow.eventOptionsWidget.setOptions()

    def on_events_loaded(self, events):
        super(PyWeedGUI, self).on_events_loaded(events)
        if self.mainWindow:
            self.mainWindow.onEventsLoaded(events)

    ###############
    # Stations
    ###############

    def set_station_options(self, options):
        super(PyWeedGUI, self).set_station_options(options)
        if self.mainWindow:
            self.mainWindow.stationOptionsWidget.setOptions()

    def on_stations_loaded(self, stations):
        super(PyWeedGUI, self).on_stations_loaded(stations)
        if self.mainWindow:
            self.mainWindow.onStationsLoaded(stations)

    ###############
    # Waveforms
    ###############

    def openWaveformsDialog(self):
        self.waveformsDialog.show()
        self.waveformsDialog.loadWaveformChoices()

    ###############
    # Summary
    ###############

    def saveSummary(self):
        """
        Save the selected events/stations
        """
        # If the user quits or cancels this dialog, '' is returned
        savePath = str(QtGui.QFileDialog.getExistingDirectory(
            parent=self.mainWindow,
            caption="Save Summary to"))
        if savePath != '':
            try:
                catalog = get_filtered_catalog(self.events, self.iter_selected_events())
                catalog.write(os.path.join(savePath, 'events.xml'), format="QUAKEML")
            except Exception as e:
                LOGGER.error("Unable to save event selection! %s", e)
            try:
                inventory = get_filtered_inventory(self.stations, self.iter_selected_stations())
                inventory.write(os.path.join(savePath, 'stations.xml'), format="STATIONXML")
            except Exception as e:
                LOGGER.error("Unable to save station selection! %s", e)

    def loadSummary(self):
        # If the user quits or cancels this dialog, '' is returned
        loadPath = str(QtGui.QFileDialog.getExistingDirectory(
            parent=self.mainWindow,
            caption="Load Summary from"))
        if loadPath != '':
            try:
                catalog = read_events(os.path.join(loadPath, 'events.xml'))
                self.on_events_loaded(catalog)
                self.mainWindow.selectAllEvents()
            except Exception as e:
                LOGGER.error("Unable to load events! %s", e)
            try:
                inventory = read_inventory(os.path.join(loadPath, 'stations.xml'))
                self.on_stations_loaded(inventory)
                self.mainWindow.selectAllStations()
            except Exception as e:
                LOGGER.error("Unable to load stations! %s", e)

    ###############
    # Other UI elements
    ###############

    def configureMenu(self):
        # Create menuBar
        # see:  http://doc.qt.io/qt-4.8/qmenubar.html
        # see:  http://zetcode.com/gui/pyqt4/menusandtoolbars/
        # see:  https://pythonprogramming.net/menubar-pyqt-tutorial/
        # see:  http://www.dreamincode.net/forums/topic/261282-a-basic-pyqt-tutorial-notepad/
        mainMenu = QtGui.QMenuBar()
        # mainMenu.setNativeMenuBar(False)

        fileMenu = mainMenu.addMenu('&File')

        saveSummaryAction = QtGui.QAction("Save Summary", self.mainWindow)
        saveSummaryAction.triggered.connect(self.saveSummary)
        fileMenu.addAction(saveSummaryAction)

        loadSummaryAction = QtGui.QAction("Load Summary", self.mainWindow)
        loadSummaryAction.triggered.connect(self.loadSummary)
        fileMenu.addAction(loadSummaryAction)

        quitAction = QtGui.QAction("&Quit", self.mainWindow)
        quitAction.setShortcut("Ctrl+Q")
        quitAction.triggered.connect(self.closeApplication)
        fileMenu.addAction(quitAction)

        viewMenu = mainMenu.addMenu('View')

        showConsoleAction = QtGui.QAction("Show Python Console", self)
        showConsoleAction.triggered.connect(self.console.show)
        viewMenu.addAction(showConsoleAction)

        showLogsAction = QtGui.QAction("Show Logs", self)
        showLogsAction.triggered.connect(self.loggingDialog.show)
        viewMenu.addAction(showLogsAction)

        optionsMenu = mainMenu.addMenu('Options')

        showPreferencesAction = QtGui.QAction("Preferences", self)
        showPreferencesAction.triggered.connect(self.preferencesDialog.exec_)
        optionsMenu.addAction(showPreferencesAction)

        helpMenu = mainMenu.addMenu('Help')

        aboutPyweedAction = QtGui.QAction("&About PYWEED", self)
        aboutPyweedAction.triggered.connect(self.showAboutDialog)
        helpMenu.addAction(aboutPyweedAction)
        helpMenu.addSeparator()

        self.mainWindow.setMenuBar(mainMenu)
        # self.waveformsDialog.setMenuBar(mainMenu)

    def showAboutDialog(self):
        """Display About message box."""
        # see:  http://www.programcreek.com/python/example/62361/PyQt4.QtGui.QMessageBox
        website = "https://github.com/iris-edu/pyweed"
        # email = "adam@iris.washington.edu"
        license_link = "https://github.com/iris-edu/pyweed/blob/master/LICENSE"
        license_name = "LGPLv3"
        mazama_link = "http://mazamascience.com"
        mazama_name = "Mazama Science"
        iris_link = "http://ds.iris.edu/ds/nodes/dmc/"
        iris_name = "IRIS"

        msgBox = QtGui.QMessageBox()
        msgBox.setWindowTitle("About %s" % __app_name__)
        msgBox.setTextFormat(QtCore.Qt.RichText)
        # msgBox.setIconPixmap(QtGui.QPixmap(ComicTaggerSettings.getGraphic('about.png')))
        msgBox.setText("<br><br><br>" +
                       __app_name__ +
                       " v" +
                       __version__ +
                       "<br><br>" +
                       "Pyweed is a cross-platform GUI application for retrieving event-based seismic data." +
                       "<br><br>" +
                       "<a href='{0}'>{0}</a><br><br>".format(website) +
                       # "<a href='mailto:{0}'>{0}</a><br><br>".format(email) +
                       "License: <a href='{0}'>{1}</a>".format(license_link, license_name) +
                       "<br><br>" +
                       "Developed by <a href='{0}'>{1}</a>".format(mazama_link, mazama_name) +
                       " for <a href='{0}'>{1}</a>".format(iris_link, iris_name) +
                       ".")

        msgBox.setStandardButtons(QtGui.QMessageBox.Ok)
        msgBox.exec_()
        # NOTE:  For info on " modalSession has been exited prematurely" error on OS X see:
        # NOTE:    https://forum.qt.io/topic/43618/modal-sessions-with-pyqt4-and-os-x/2

    def closeApplication(self):
        LOGGER.info('Closing application...')
        # Update preferences
        self.mainWindow.savePreferences()
        self.waveformsDialog.savePreferences()
        self.close()
        QtGui.QApplication.quit()


if __name__ == "__main__":
    print("Run pyweed_launcher.py instead!")



# -*- coding: utf-8 -*-
"""
Station Options

:copyright:
    Mazama Science, IRIS
:license:
    GNU Lesser General Public License, Version 3
    (http://www.gnu.org/copyleft/lesser.html)
"""

from pyweed.options import Options, DateOption, FloatOption, Option
import copy
from pyweed.dist_from_events import get_combined_locations, CrossBorderException
from pyweed.pyweed_utils import get_preferred_origin
from logging import getLogger

LOGGER = getLogger(__name__)


class StationOptions(Options):

    TIME_RANGE = 'timeBetween'
    TIME_EVENTS = 'timeFromEvents'
    time_choice = Option(hidden=True, default=TIME_RANGE)

    starttime = DateOption(default=-30)
    endtime = DateOption(default=0)

    network = Option(default="_GSN")
    station = Option(default="*")
    location = Option(default="*")
    channel = Option(default="?HZ")

    LOCATION_GLOBAL = 'locationGlobal'
    LOCATION_BOX = 'locationRange'
    LOCATION_POINT = 'locationDistanceFromPoint'
    LOCATION_EVENTS = 'locationDistanceFromEvents'
    location_choice = Option(hidden=True, default=LOCATION_GLOBAL)

    minlatitude = FloatOption(default=-90)
    maxlatitude = FloatOption(default=90)
    minlongitude = FloatOption(default=-180)
    maxlongitude = FloatOption(default=180)

    latitude = FloatOption()
    longitude = FloatOption()
    minradius = FloatOption()
    maxradius = FloatOption(default=30)

    # Special settings indicating the location is based on distance from selected events
    mindistance = FloatOption(hidden=True)
    maxdistance = FloatOption(hidden=True)

    event_locations = None

    def get_time_options(self):
        if self.time_choice == self.TIME_RANGE:
            return self.get_options(['starttime', 'endtime'])
        else:
            return {}

    def get_location_options(self):
        if self.location_choice == self.LOCATION_BOX:
            return self.get_options(['minlatitude', 'maxlatitude', 'minlongitude', 'maxlongitude'])
        elif self.location_choice == self.LOCATION_POINT:
            return self.get_options(['latitude', 'longitude', 'minradius', 'maxradius'])
        else:
            return {}

    def get_obspy_options(self):
        base_keys = ['network', 'station', 'location', 'channel']
        options = self.get_options(keys=base_keys)
        # Default level is 'station' but we always(?) want 'channel'
        options['level'] = 'channel'
        options.update(self.get_time_options())
        options.update(self.get_location_options())
        return options

    def get_event_distances(self):
        """
        If this is a event-based query, return those settings
        Note that this isn't part of the Obspy query!
        """
        if self.location_choice == self.LOCATION_EVENTS:
            return self.get_options(['mindistance', 'maxdistance'])
        else:
            return {}





# -*- coding: utf-8 -*-
"""
Preferences dialog

:copyright:
    Mazama Science, IRIS
:license:
    GNU Lesser General Public License, Version 3
    (http://www.gnu.org/copyleft/lesser.html)
"""

from PyQt4 import QtGui
from pyweed.gui.uic import PreferencesDialog
from logging import getLogger
from obspy.clients.fdsn.header import URL_MAPPINGS
from obspy.clients.fdsn.client import Client

LOGGER = getLogger(__name__)


class PreferencesDialog(QtGui.QDialog, PreferencesDialog.Ui_PreferencesDialog):
    """
    Dialog window for editing preferences.
    """
    def __init__(self, pyweed, parent=None):
        super(PreferencesDialog, self).__init__(parent=parent)
        self.setupUi(self)

        self.pyweed = pyweed

        # Ordered list of all available data centers
        self.data_centers = sorted(URL_MAPPINGS.keys())

        # Put these in the comboboxes
        for data_center in self.data_centers:
            label = "%s: %s" % (data_center, URL_MAPPINGS[data_center])
            self.eventDataCenterComboBox.addItem(label, data_center)
            self.stationDataCenterComboBox.addItem(label, data_center)

        self.okButton.pressed.connect(self.accept)
        self.cancelButton.pressed.connect(self.reject)

    def showEvent(self, *args, **kwargs):
        """
        Perform any necessary initialization each time the dialog is opened
        """
        super(PreferencesDialog, self).showEvent(*args, **kwargs)
        # Indicate the currently selected data centers
        self.eventDataCenterComboBox.setCurrentIndex(self.data_centers.index(self.pyweed.event_data_center))
        self.stationDataCenterComboBox.setCurrentIndex(self.data_centers.index(self.pyweed.station_data_center))
        self.cacheSizeSpinBox.setValue(int(self.pyweed.preferences.Waveforms.cacheSize))

    def accept(self):
        """
        Validate and update the client
        """
        try:
            self.pyweed.set_event_data_center(
                self.data_centers[self.eventDataCenterComboBox.currentIndex()])
            self.pyweed.set_station_data_center(
                self.data_centers[self.stationDataCenterComboBox.currentIndex()])
        except Exception as e:
            # Error usually means that the user selected a data center that doesn't provide the given service
            QtGui.QMessageBox.critical(
                self,
                "Unable to update data center",
                str(e)
            )
            # Don't call super() which would close the preferences dialog
            return

        self.pyweed.preferences.Waveforms.cacheSize = self.cacheSizeSpinBox.value()

        return super(PreferencesDialog, self).accept()



# -*- coding: utf-8 -*-
"""
Widgets for use with a QTableWidgetItem

:copyright:
    Mazama Science, IRIS
:license:
    GNU Lesser General Public License, Version 3
    (http://www.gnu.org/copyleft/lesser.html)
"""

from PyQt4 import QtGui, QtCore

# QTableWidgetItem subclass type iterator
_next_type = QtGui.QTableWidgetItem.UserType


class CustomTableWidgetItemMixin(object):
    def __init__(self, *args, **kwargs):
        global _next_type
        _next_type += 1
        return super(CustomTableWidgetItemMixin, self).__init__(*args, **kwargs, type=_next_type)



# -*- coding: utf-8 -*-

# Form implementation generated from reading ui file 'pyweed/gui/uic/LoggingDialog.ui'
#
# Created by: PyQt4 UI code generator 4.11.4
#
# WARNING! All changes made in this file will be lost!

from PyQt4 import QtCore, QtGui

try:
    _fromUtf8 = QtCore.QString.fromUtf8
except AttributeError:
    def _fromUtf8(s):
        return s

try:
    _encoding = QtGui.QApplication.UnicodeUTF8
    def _translate(context, text, disambig):
        return QtGui.QApplication.translate(context, text, disambig, _encoding)
except AttributeError:
    def _translate(context, text, disambig):
        return QtGui.QApplication.translate(context, text, disambig)

class Ui_LoggingDialog(object):
    def setupUi(self, LoggingDialog):
        LoggingDialog.setObjectName(_fromUtf8("LoggingDialog"))
        LoggingDialog.resize(697, 474)
        self.verticalLayout = QtGui.QVBoxLayout(LoggingDialog)
        self.verticalLayout.setObjectName(_fromUtf8("verticalLayout"))
        self.loggingPlainTextEdit = QtGui.QPlainTextEdit(LoggingDialog)
        self.loggingPlainTextEdit.setObjectName(_fromUtf8("loggingPlainTextEdit"))
        self.verticalLayout.addWidget(self.loggingPlainTextEdit)

        self.retranslateUi(LoggingDialog)
        QtCore.QMetaObject.connectSlotsByName(LoggingDialog)

    def retranslateUi(self, LoggingDialog):
        LoggingDialog.setWindowTitle(_translate("LoggingDialog", "Dialog", None))




# -*- coding: utf-8 -*-

# Form implementation generated from reading ui file 'pyweed/gui/uic/SpinnerWidget.ui'
#
# Created by: PyQt4 UI code generator 4.11.4
#
# WARNING! All changes made in this file will be lost!

from PyQt4 import QtCore, QtGui

try:
    _fromUtf8 = QtCore.QString.fromUtf8
except AttributeError:
    def _fromUtf8(s):
        return s

try:
    _encoding = QtGui.QApplication.UnicodeUTF8
    def _translate(context, text, disambig):
        return QtGui.QApplication.translate(context, text, disambig, _encoding)
except AttributeError:
    def _translate(context, text, disambig):
        return QtGui.QApplication.translate(context, text, disambig)

class Ui_SpinnerWidget(object):
    def setupUi(self, SpinnerWidget):
        SpinnerWidget.setObjectName(_fromUtf8("SpinnerWidget"))
        SpinnerWidget.resize(306, 207)
        sizePolicy = QtGui.QSizePolicy(QtGui.QSizePolicy.Ignored, QtGui.QSizePolicy.Ignored)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(SpinnerWidget.sizePolicy().hasHeightForWidth())
        SpinnerWidget.setSizePolicy(sizePolicy)
        SpinnerWidget.setStyleSheet(_fromUtf8("QFrame { background-color: rgba(224,224,224,192)} \n"
"QLabel { background-color: transparent }"))
        self.verticalLayout = QtGui.QVBoxLayout(SpinnerWidget)
        self.verticalLayout.setSpacing(0)
        self.verticalLayout.setObjectName(_fromUtf8("verticalLayout"))
        self.icon = QtGui.QLabel(SpinnerWidget)
        self.icon.setText(_fromUtf8(""))
        self.icon.setAlignment(QtCore.Qt.AlignBottom|QtCore.Qt.AlignHCenter)
        self.icon.setObjectName(_fromUtf8("icon"))
        self.verticalLayout.addWidget(self.icon)
        self.label = QtGui.QLabel(SpinnerWidget)
        self.label.setText(_fromUtf8(""))
        self.label.setAlignment(QtCore.Qt.AlignHCenter|QtCore.Qt.AlignTop)
        self.label.setObjectName(_fromUtf8("label"))
        self.verticalLayout.addWidget(self.label)
        self.horizontalLayout = QtGui.QHBoxLayout()
        self.horizontalLayout.setObjectName(_fromUtf8("horizontalLayout"))
        self.cancelButton = QtGui.QPushButton(SpinnerWidget)
        sizePolicy = QtGui.QSizePolicy(QtGui.QSizePolicy.Fixed, QtGui.QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.cancelButton.sizePolicy().hasHeightForWidth())
        self.cancelButton.setSizePolicy(sizePolicy)
        self.cancelButton.setObjectName(_fromUtf8("cancelButton"))
        self.horizontalLayout.addWidget(self.cancelButton)
        self.verticalLayout.addLayout(self.horizontalLayout)
        self.verticalLayout.setStretch(0, 1)
        self.verticalLayout.setStretch(1, 1)

        self.retranslateUi(SpinnerWidget)
        QtCore.QMetaObject.connectSlotsByName(SpinnerWidget)

    def retranslateUi(self, SpinnerWidget):
        SpinnerWidget.setWindowTitle(_translate("SpinnerWidget", "Form", None))
        self.cancelButton.setText(_translate("SpinnerWidget", "Cancel", None))




# -*- coding: utf-8 -*-
"""
Event Options

:copyright:
    Mazama Science, IRIS
:license:
    GNU Lesser General Public License, Version 3
    (http://www.gnu.org/copyleft/lesser.html)
"""

from PyQt4 import QtGui, QtCore
from pyweed.gui.uic import EventOptionsWidget
import logging
from distutils.util import strtobool
from pyweed.event_options import EventOptions
from pyweed.gui.OptionsWidget import BaseOptionsWidget

LOGGER = logging.getLogger(__name__)


class EventOptionsWidget(BaseOptionsWidget, EventOptionsWidget.Ui_EventOptionsWidget):
    """
    Dialog window for event options used in creating a webservice query.
    """

    def mapInputs(self):
        return {
            'starttime': self.starttimeDateTimeEdit,
            'endtime': self.endtimeDateTimeEdit,
            'minmagnitude': self.minMagDoubleSpinBox,
            'maxmagnitude': self.maxMagDoubleSpinBox,
            'mindepth': self.minDepthDoubleSpinBox,
            'maxdepth': self.maxDepthDoubleSpinBox,
            'magnitudetype': self.magTypeComboBox,
            'minlongitude': self.locationRangeWestDoubleSpinBox,
            'maxlongitude': self.locationRangeEastDoubleSpinBox,
            'minlatitude': self.locationRangeSouthDoubleSpinBox,
            'maxlatitude': self.locationRangeNorthDoubleSpinBox,
            'minradius': self.distanceFromPointMinRadiusDoubleSpinBox,
            'maxradius': self.distanceFromPointMaxRadiusDoubleSpinBox,
            'longitude': self.distanceFromPointEastDoubleSpinBox,
            'latitude': self.distanceFromPointNorthDoubleSpinBox,
            '_locationGlobal': self.locationGlobalRadioButton,
            '_locationRange': self.locationRangeRadioButton,
            '_locationDistanceFromPoint': self.locationDistanceFromPointRadioButton
        }

    def optionsToInputs(self, values):
        # Turn the option choice value into a set of radio values
        location_choice = values.get('location_choice')
        values['_locationGlobal'] = (location_choice == EventOptions.LOCATION_GLOBAL)
        values['_locationRange'] = (location_choice == EventOptions.LOCATION_BOX)
        values['_locationDistanceFromPoint'] = (location_choice == EventOptions.LOCATION_POINT)
        return values

    def inputsToOptions(self, values):
        # Turn the radio values into an option choice value
        if strtobool(values.get('_locationGlobal', 'F')):
            values['location_choice'] = EventOptions.LOCATION_GLOBAL
        elif strtobool(values.get('_locationRange', 'F')):
            values['location_choice'] = EventOptions.LOCATION_BOX
        elif strtobool(values.get('_locationDistanceFromPoint', 'F')):
            values['location_choice'] = EventOptions.LOCATION_POINT
        return values

    def get_timeFromOtherButton(self):
        return self.timeFromStationsToolButton

    def get_locationFromOtherButton(self):
        return self.locationFromStationsToolButton




# -*- coding: utf-8 -*-
"""
Base classes standardizing some Qt signals.

:copyright:
    Mazama Science, IRIS
:license:
    GNU Lesser General Public License, Version 3
    (http://www.gnu.org/copyleft/lesser.html)
"""

from PyQt4 import QtCore


# NOTE:  http://stackoverflow.com/questions/9957195/updating-gui-elements-in-multithreaded-pyqt
class SignalingThread(QtCore.QThread):
    """
    Mixin that serves to standardize how we use signals to pass information around
    """
    # Signal to indicate success
    done = QtCore.pyqtSignal(object)

    def __del__(self):
        self.wait()


class SignalingObject(QtCore.QObject):
    """
    Mixin for other objects that manage threaded work, and need to signal completion
    """
    # Signal to indicate success
    done = QtCore.pyqtSignal(object)



# -*- coding: utf-8 -*-
"""
Pyweed utility functions.

:copyright:
    Mazama Science, IRIS
:license:
    GNU Lesser General Public License, Version 3
    (http://www.gnu.org/copyleft/lesser.html)
"""

from __future__ import (absolute_import, division, print_function)

# Basic packages
import os
import logging
import re
from pyproj import Geod
from obspy.taup.tau import TauPyModel
from future.moves.urllib.parse import urlencode

LOGGER = logging.getLogger(__name__)
GEOD = Geod(ellps='WGS84')
TAUP = TauPyModel()

# Rough meters/degree calculation
M_PER_DEG = (GEOD.inv(0, 0, 0, 1)[2] + GEOD.inv(0, 0, 1, 0)[2]) / 2



class OutputFormat(object):
    """
    Simple output format definition.
    """
    def __init__(self, value, label=None, extension=None):
        #: This is the name used by ObsPy, which we treat as the "real" value
        self.value = value
        #: This is the label used in the UI, it defaults to value
        self.label = label or value
        #: This is the file extension, it defaults to lowercased value
        self.extension = extension or value.lower()

# List of the output formats we support
OUTPUT_FORMATS = [
    OutputFormat('MSEED', 'MiniSEED'),
    OutputFormat('SAC'),
    OutputFormat('SACXY', 'SAC ASCII', 'sac.txt'),
    OutputFormat('SLIST', 'ASCII (1 column)', 'ascii1.txt'),
    OutputFormat('TSPAIR', 'ASCII (2 column)', 'ascii2.txt'),
]
# Map of a format to the file extension to use
OUTPUT_FORMAT_EXTENSIONS = dict(((f.value, f.extension) for f in OUTPUT_FORMATS))


class Phase(object):
    """
    Simple phase definition
    """
    def __init__(self, name, label):
        self.name = name
        self.label = label

# List of phases we can use for time windows
EVENT_TIME_PHASE = 'Event'
PHASES = [
    Phase('P', 'P wave arrival'),
    Phase('S', 'S wave arrival'),
    Phase(EVENT_TIME_PHASE, 'Event time')
]
# Actual phase values retrieved from TauP, this should give us a good P and S value for any input (I hope!)
TAUP_PHASES = ['P', 'PKIKP', 'Pdiff', 'S', 'SKIKS', 'SKS', 'p', 's']


def manage_cache(download_dir, cache_size):
    """
    Maintain a cache directory at a certain size (MB) by removing the oldest files.
    """

    try:
        # Compile statistics on all files in the output directory and subdirectories
        stats = []
        total_size = 0
        for root, dirs, files in os.walk(download_dir):
            for file in files:
                path = os.path.join(root, file)
                stat_list = os.stat(path)
                # path, size, atime
                new_stat_list = [path, stat_list.st_size, stat_list.st_atime]
                total_size = total_size + stat_list.st_size
                # don't want hidden files like .htaccess so don't add stuff that starts with .
                if not file.startswith('.'):
                    stats.append(new_stat_list)

        # Sort file stats by last access time
        stats = sorted(stats, key=lambda file: file[2])

        # Delete old files until we get under cache_size (configured in megabytes)
        deletion_count = 0
        while total_size > cache_size * 1000000:
            # front of stats list is the file with the smallest (=oldest) access time
            last_accessed_file = stats[0]
            # index 1 is where size is
            total_size = total_size - last_accessed_file[1]
            # index 0 is where path is
            os.remove(last_accessed_file[0])
            # remove the file from the stats list
            stats.pop(0)
            deletion_count = deletion_count + 1

        LOGGER.debug('Removed %d files to keep %s below %.0f megabytes' % (deletion_count, download_dir, cache_size))

    except Exception as e:
        LOGGER.error(str(e))


def iter_channels(inventory, dedupe=True):
    """
    Iterate over every channel in an inventory.
    For each channel, yields (network, station, channel)

    If dedupe=True, repeated channels are filtered out -- this can occur if the inventory includes
    multiple epochs for a given channel. Only the first channel will be included in this case.
    """
    last_sncl = None
    if inventory:
        for network in inventory.networks:
            for station in network.stations:
                for channel in station.channels:
                    if dedupe:
                        sncl = get_sncl(network, station, channel)
                        if sncl == last_sncl:
                            continue
                        last_sncl = sncl
                    yield (network, station, channel)


def get_sncl(network, station, channel):
    """
    Generate the SNCL for the given network/station/channel
    """
    return '.'.join((network.code, station.code, channel.location_code, channel.code))


def get_event_id(event):
    """
    Get a unique ID for a given event

    Event IDs are given differently by different data centers.

    Examples compiled by crotwell@seis.sc.edu:

    IRIS
    <event publicID="smi:service.iris.edu/fdsnws/event/1/query?eventid=3337497">

    NCEDC
    <event publicID="quakeml:nc.anss.org/Event/NC/71377596"
    catalog:datasource="nc" catalog:dataid="nc71377596"
    catalog:eventsource="nc" catalog:eventid="71377596">

    SCEDC
    <event publicID="quakeml:service.scedc.caltech.edu/fdsnws/event/1/query?eventid=37300872"
    catalog:datasource="ci" catalog:dataid="ci37300872"
    catalog:eventsource="ci" catalog:eventid="37300872">

    USGS
    <event catalog:datasource="us" catalog:eventsource="us"
    catalog:eventid="c000lvb5"
    publicID="quakeml:earthquake.usgs.gov/fdsnws/event/1/query?eventid=usc000lvb5&format=quakeml">

    ETHZ
    <event publicID="smi:ch.ethz.sed/sc3a/2017eemfch">

    INGV
    <event publicID="smi:webservices.ingv.it/fdsnws/event/1/query?eventId=863301">

    ISC
    <event publicID="smi:ISC/evid=600516598">
    """
    # Look for "eventid=" as a URL query parameter
    m = re.search(r'eventid=([^\&]+)', event.resource_id.id, re.IGNORECASE)
    if m:
        return m.group(1)
    # Otherwise, return the trailing segment of alphanumerics
    return re.sub(r'^.*?(\w+)\W*$', r'\1', event.resource_id.id)


def get_event_name(event):
    time_str = get_event_time_str(event)
    mag_str = get_event_mag_str(event)
    description = get_event_description(event)
    return "%s | %s | %s" % (time_str, mag_str, description)


def format_time_str(time):
    return time.isoformat(sep=' ').split('.')[0]


def get_event_time_str(event):
    origin = get_preferred_origin(event)
    return format_time_str(origin.time)


def get_event_mag_str(event):
    mag = get_preferred_magnitude(event)
    return "%s%s" % (mag.mag, mag.magnitude_type)


def get_event_description(event):
    return str(event.event_descriptions[0].text).title()


def get_preferred_origin(event):
    """
    Get the preferred origin for the event, or None if not defined
    """
    origin = event.preferred_origin()
    if not origin:
        LOGGER.error("No preferred origin found for event %s", event.resource_id)
        if len(event.origins):
            origin = event.origins[0]
    return origin


def get_preferred_magnitude(event):
    """
    Get the preferred magnitude for the event, or None if not defined
    """
    magnitude = event.preferred_magnitude()
    if not magnitude:
        LOGGER.error("No preferred magnitude found for event %s", event.resource_id)
        if len(event.magnitudes):
            magnitude = event.magnitudes[0]
    return magnitude


class TimeWindow(object):
    """
    Represents a time window for data based on phase arrivals at a particular location
    """
    start_offset = 0
    end_offset = 0
    start_phase = None
    end_phase = None

    def __init__(self, start_offset=0, end_offset=0, start_phase=PHASES[0].name, end_phase=PHASES[0].name):
        self.update(start_offset, end_offset, start_phase, end_phase)

    def update(self, start_offset, end_offset, start_phase, end_phase):
        """
        Set all values. Phases can be specified by name or label.
        """
        self.start_offset = start_offset
        self.end_offset = end_offset
        self.start_phase = start_phase
        self.end_phase = end_phase

    def calculate_window(self, event_time, arrivals):
        """
        Given an event time and a dictionary of arrival times (see Distances below)
        calculate the full time window
        """
        return (
            # Start time
            event_time + arrivals.get(self.start_phase, 0) - self.start_offset,
            # End time
            event_time + arrivals.get(self.end_phase, 0) + self.end_offset,
        )

    def __eq__(self, other):
        """
        Compare two TimeWindows
        """
        if not isinstance(other, TimeWindow):
            return False
        return (other.start_offset == self.start_offset and
                other.end_offset == self.end_offset and
                other.start_phase == self.start_phase and
                other.end_phase == self.end_phase)


def get_distance(lat1, lon1, lat2, lon2):
    """
    Get the distance between two points in degrees
    """
    # NOTE that GEOD takes longitude first!
    _az, _baz, meters = GEOD.inv(lon1, lat1, lon2, lat2)
    return meters / M_PER_DEG


def get_arrivals(distance, event_depth):
    """
    Calculate phase arrival times

    :param distance: distance in degrees
    :param event_depth: event depth in km
    """
    arrivals = TAUP.get_travel_times(
        event_depth,
        distance,
        TAUP_PHASES
    )

    # From the travel time and origin, calculate the actual first arrival time for each basic phase type
    first_arrivals = {}
    for arrival in arrivals:
        # The basic phase name is the uppercase first letter of the full phase name
        # We assume this matches a Phase.name defined in PHASES
        phase_name = arrival.name[0].upper()
        if phase_name not in first_arrivals:
            first_arrivals[phase_name] = arrival.time

    return first_arrivals


def get_bounding_circle(lat, lon, radius, num_points=36):
    """
    Returns groups of lat/lon pairs representing a circle on the map
    """
    radius_meters = radius * M_PER_DEG
    # NOTE that GEOD takes longitude first!
    trans = GEOD.fwd(
        [lon] * num_points,
        [lat] * num_points,
        list(((i * 360) / num_points) for i in range(num_points)),
        [radius_meters] * num_points
    )
    points = list(zip(trans[1], trans[0]))
    # We need to complete the circle by adding the first point again as the last point
    points.append(points[0])
    return points


def get_service_url(client, service, parameters):
    """
    Figure out the URL for the given service call. This isn't publicly available from the ObsPy client,
    we need to use internal APIs, so those messy details are encapsulated here.
    """
    try:
        return client._create_url_from_parameters(service, {}, parameters)
    except:
        return "%s %s %s" % (
            client.base_url, service, urlencode(parameters)
        )


class CancelledException(Exception):
    """
    An exception to return in an event notification indicating that an operation was cancelled.
    See `StationsHandler` for an example.
    """
    def __str__(self, *args, **kwargs):
        s = super(CancelledException, self).__str__(*args, **kwargs)
        if s == '':
            return 'Cancelled'
        else:
            return s


class DataRequest(object):
    """
    Wrapper object for a data request, which may or may not be more than a single web service query.
    """
    # The client to use
    client = None
    # List of option dictionaries, one for each sub-request required
    sub_requests = None

    def __init__(self, client, *requests):
        self.client = client
        self.sub_requests = requests

    def process_result(self, result):
        """
        Subclasses can define behavior here to do post-processing on the resulting data
        """
        return result



# -*- coding: utf-8 -*-

# Form implementation generated from reading ui file 'pyweed/gui/uic/MainWindow.ui'
#
# Created by: PyQt4 UI code generator 4.11.4
#
# WARNING! All changes made in this file will be lost!

from PyQt4 import QtCore, QtGui

try:
    _fromUtf8 = QtCore.QString.fromUtf8
except AttributeError:
    def _fromUtf8(s):
        return s

try:
    _encoding = QtGui.QApplication.UnicodeUTF8
    def _translate(context, text, disambig):
        return QtGui.QApplication.translate(context, text, disambig, _encoding)
except AttributeError:
    def _translate(context, text, disambig):
        return QtGui.QApplication.translate(context, text, disambig)

class Ui_MainWindow(object):
    def setupUi(self, MainWindow):
        MainWindow.setObjectName(_fromUtf8("MainWindow"))
        MainWindow.resize(1037, 594)
        MainWindow.setStyleSheet(_fromUtf8("QTableView { gridline-color: gray }"))
        self.centralwidget = QtGui.QWidget(MainWindow)
        self.centralwidget.setObjectName(_fromUtf8("centralwidget"))
        self.verticalLayout = QtGui.QVBoxLayout(self.centralwidget)
        self.verticalLayout.setContentsMargins(-1, -1, -1, 0)
        self.verticalLayout.setObjectName(_fromUtf8("verticalLayout"))
        self.splitter = QtGui.QSplitter(self.centralwidget)
        self.splitter.setOrientation(QtCore.Qt.Vertical)
        self.splitter.setChildrenCollapsible(False)
        self.splitter.setObjectName(_fromUtf8("splitter"))
        self.mapFrame = QtGui.QFrame(self.splitter)
        self.mapFrame.setFrameShape(QtGui.QFrame.StyledPanel)
        self.mapFrame.setFrameShadow(QtGui.QFrame.Raised)
        self.mapFrame.setObjectName(_fromUtf8("mapFrame"))
        self.verticalLayout_4 = QtGui.QVBoxLayout(self.mapFrame)
        self.verticalLayout_4.setMargin(0)
        self.verticalLayout_4.setSpacing(0)
        self.verticalLayout_4.setObjectName(_fromUtf8("verticalLayout_4"))
        self.mapCanvas = MyQt4MplCanvas(self.mapFrame)
        self.mapCanvas.setObjectName(_fromUtf8("mapCanvas"))
        self.verticalLayout_4.addWidget(self.mapCanvas)
        self.horizontalLayout_3 = QtGui.QHBoxLayout()
        self.horizontalLayout_3.setSpacing(2)
        self.horizontalLayout_3.setObjectName(_fromUtf8("horizontalLayout_3"))
        spacerItem = QtGui.QSpacerItem(0, 0, QtGui.QSizePolicy.Expanding, QtGui.QSizePolicy.Minimum)
        self.horizontalLayout_3.addItem(spacerItem)
        self.mapZoomInButton = QtGui.QToolButton(self.mapFrame)
        sizePolicy = QtGui.QSizePolicy(QtGui.QSizePolicy.Fixed, QtGui.QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.mapZoomInButton.sizePolicy().hasHeightForWidth())
        self.mapZoomInButton.setSizePolicy(sizePolicy)
        self.mapZoomInButton.setObjectName(_fromUtf8("mapZoomInButton"))
        self.horizontalLayout_3.addWidget(self.mapZoomInButton)
        self.mapZoomOutButton = QtGui.QToolButton(self.mapFrame)
        sizePolicy = QtGui.QSizePolicy(QtGui.QSizePolicy.Fixed, QtGui.QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.mapZoomOutButton.sizePolicy().hasHeightForWidth())
        self.mapZoomOutButton.setSizePolicy(sizePolicy)
        self.mapZoomOutButton.setObjectName(_fromUtf8("mapZoomOutButton"))
        self.horizontalLayout_3.addWidget(self.mapZoomOutButton)
        self.mapResetButton = QtGui.QToolButton(self.mapFrame)
        sizePolicy = QtGui.QSizePolicy(QtGui.QSizePolicy.Fixed, QtGui.QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.mapResetButton.sizePolicy().hasHeightForWidth())
        self.mapResetButton.setSizePolicy(sizePolicy)
        self.mapResetButton.setObjectName(_fromUtf8("mapResetButton"))
        self.horizontalLayout_3.addWidget(self.mapResetButton)
        spacerItem1 = QtGui.QSpacerItem(0, 0, QtGui.QSizePolicy.Expanding, QtGui.QSizePolicy.Minimum)
        self.horizontalLayout_3.addItem(spacerItem1)
        self.verticalLayout_4.addLayout(self.horizontalLayout_3)
        self.eventAndStationWidget = QtGui.QWidget(self.splitter)
        self.eventAndStationWidget.setObjectName(_fromUtf8("eventAndStationWidget"))
        self.verticalLayout_3 = QtGui.QVBoxLayout(self.eventAndStationWidget)
        self.verticalLayout_3.setMargin(0)
        self.verticalLayout_3.setObjectName(_fromUtf8("verticalLayout_3"))
        self.splitter_2 = QtGui.QSplitter(self.eventAndStationWidget)
        self.splitter_2.setFrameShape(QtGui.QFrame.NoFrame)
        self.splitter_2.setOrientation(QtCore.Qt.Horizontal)
        self.splitter_2.setChildrenCollapsible(False)
        self.splitter_2.setObjectName(_fromUtf8("splitter_2"))
        self.eventsWidget = QtGui.QWidget(self.splitter_2)
        self.eventsWidget.setObjectName(_fromUtf8("eventsWidget"))
        self.eventsLayout = QtGui.QVBoxLayout(self.eventsWidget)
        self.eventsLayout.setMargin(0)
        self.eventsLayout.setSpacing(0)
        self.eventsLayout.setObjectName(_fromUtf8("eventsLayout"))
        self.eventsButtonLayout = QtGui.QHBoxLayout()
        self.eventsButtonLayout.setSpacing(12)
        self.eventsButtonLayout.setObjectName(_fromUtf8("eventsButtonLayout"))
        self.toggleEventOptions = QtGui.QToolButton(self.eventsWidget)
        sizePolicy = QtGui.QSizePolicy(QtGui.QSizePolicy.Fixed, QtGui.QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.toggleEventOptions.sizePolicy().hasHeightForWidth())
        self.toggleEventOptions.setSizePolicy(sizePolicy)
        self.toggleEventOptions.setCheckable(True)
        self.toggleEventOptions.setChecked(True)
        self.toggleEventOptions.setObjectName(_fromUtf8("toggleEventOptions"))
        self.eventsButtonLayout.addWidget(self.toggleEventOptions)
        self.getEventsButton = QtGui.QPushButton(self.eventsWidget)
        self.getEventsButton.setDefault(True)
        self.getEventsButton.setObjectName(_fromUtf8("getEventsButton"))
        self.eventsButtonLayout.addWidget(self.getEventsButton)
        self.eventsLayout.addLayout(self.eventsButtonLayout)
        self.eventsTable = QtGui.QTableWidget(self.eventsWidget)
        self.eventsTable.setEditTriggers(QtGui.QAbstractItemView.NoEditTriggers)
        self.eventsTable.setAlternatingRowColors(True)
        self.eventsTable.setSelectionMode(QtGui.QAbstractItemView.MultiSelection)
        self.eventsTable.setSelectionBehavior(QtGui.QAbstractItemView.SelectRows)
        self.eventsTable.setShowGrid(False)
        self.eventsTable.setWordWrap(False)
        self.eventsTable.setObjectName(_fromUtf8("eventsTable"))
        self.eventsTable.setColumnCount(0)
        self.eventsTable.setRowCount(0)
        self.eventsTable.verticalHeader().setVisible(False)
        self.eventsLayout.addWidget(self.eventsTable)
        self.horizontalLayout = QtGui.QHBoxLayout()
        self.horizontalLayout.setObjectName(_fromUtf8("horizontalLayout"))
        self.eventSelectionLabel = QtGui.QLabel(self.eventsWidget)
        self.eventSelectionLabel.setText(_fromUtf8(""))
        self.eventSelectionLabel.setObjectName(_fromUtf8("eventSelectionLabel"))
        self.horizontalLayout.addWidget(self.eventSelectionLabel)
        self.clearEventSelectionButton = QtGui.QToolButton(self.eventsWidget)
        self.clearEventSelectionButton.setObjectName(_fromUtf8("clearEventSelectionButton"))
        self.horizontalLayout.addWidget(self.clearEventSelectionButton)
        self.eventsLayout.addLayout(self.horizontalLayout)
        self.stationsWidget = QtGui.QWidget(self.splitter_2)
        self.stationsWidget.setObjectName(_fromUtf8("stationsWidget"))
        self.stationsLayout = QtGui.QVBoxLayout(self.stationsWidget)
        self.stationsLayout.setMargin(0)
        self.stationsLayout.setSpacing(0)
        self.stationsLayout.setObjectName(_fromUtf8("stationsLayout"))
        self.stationsButtonLayout = QtGui.QHBoxLayout()
        self.stationsButtonLayout.setSpacing(12)
        self.stationsButtonLayout.setObjectName(_fromUtf8("stationsButtonLayout"))
        self.getStationsButton = QtGui.QPushButton(self.stationsWidget)
        self.getStationsButton.setDefault(True)
        self.getStationsButton.setObjectName(_fromUtf8("getStationsButton"))
        self.stationsButtonLayout.addWidget(self.getStationsButton)
        self.toggleStationOptions = QtGui.QToolButton(self.stationsWidget)
        sizePolicy = QtGui.QSizePolicy(QtGui.QSizePolicy.Fixed, QtGui.QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.toggleStationOptions.sizePolicy().hasHeightForWidth())
        self.toggleStationOptions.setSizePolicy(sizePolicy)
        self.toggleStationOptions.setCheckable(True)
        self.toggleStationOptions.setChecked(True)
        self.toggleStationOptions.setObjectName(_fromUtf8("toggleStationOptions"))
        self.stationsButtonLayout.addWidget(self.toggleStationOptions)
        self.stationsLayout.addLayout(self.stationsButtonLayout)
        self.stationsTable = QtGui.QTableWidget(self.stationsWidget)
        self.stationsTable.setEditTriggers(QtGui.QAbstractItemView.NoEditTriggers)
        self.stationsTable.setAlternatingRowColors(True)
        self.stationsTable.setSelectionMode(QtGui.QAbstractItemView.MultiSelection)
        self.stationsTable.setSelectionBehavior(QtGui.QAbstractItemView.SelectRows)
        self.stationsTable.setShowGrid(False)
        self.stationsTable.setWordWrap(False)
        self.stationsTable.setObjectName(_fromUtf8("stationsTable"))
        self.stationsTable.setColumnCount(0)
        self.stationsTable.setRowCount(0)
        self.stationsTable.verticalHeader().setVisible(False)
        self.stationsLayout.addWidget(self.stationsTable)
        self.horizontalLayout_2 = QtGui.QHBoxLayout()
        self.horizontalLayout_2.setObjectName(_fromUtf8("horizontalLayout_2"))
        self.stationSelectionLabel = QtGui.QLabel(self.stationsWidget)
        self.stationSelectionLabel.setText(_fromUtf8(""))
        self.stationSelectionLabel.setObjectName(_fromUtf8("stationSelectionLabel"))
        self.horizontalLayout_2.addWidget(self.stationSelectionLabel)
        self.clearStationSelectionButton = QtGui.QToolButton(self.stationsWidget)
        self.clearStationSelectionButton.setObjectName(_fromUtf8("clearStationSelectionButton"))
        self.horizontalLayout_2.addWidget(self.clearStationSelectionButton)
        self.stationsLayout.addLayout(self.horizontalLayout_2)
        self.verticalLayout_3.addWidget(self.splitter_2)
        self.waveformButtonLayout = QtGui.QHBoxLayout()
        self.waveformButtonLayout.setObjectName(_fromUtf8("waveformButtonLayout"))
        self.getWaveformsButton = QtGui.QPushButton(self.eventAndStationWidget)
        self.getWaveformsButton.setDefault(True)
        self.getWaveformsButton.setObjectName(_fromUtf8("getWaveformsButton"))
        self.waveformButtonLayout.addWidget(self.getWaveformsButton)
        self.verticalLayout_3.addLayout(self.waveformButtonLayout)
        self.verticalLayout_3.setStretch(0, 1)
        self.verticalLayout.addWidget(self.splitter)
        MainWindow.setCentralWidget(self.centralwidget)
        self.eventOptionsDockWidget = QtGui.QDockWidget(MainWindow)
        self.eventOptionsDockWidget.setFloating(False)
        self.eventOptionsDockWidget.setFeatures(QtGui.QDockWidget.AllDockWidgetFeatures)
        self.eventOptionsDockWidget.setAllowedAreas(QtCore.Qt.LeftDockWidgetArea|QtCore.Qt.RightDockWidgetArea)
        self.eventOptionsDockWidget.setObjectName(_fromUtf8("eventOptionsDockWidget"))
        self.dockWidgetContents = QtGui.QWidget()
        self.dockWidgetContents.setObjectName(_fromUtf8("dockWidgetContents"))
        self.eventOptionsDockWidget.setWidget(self.dockWidgetContents)
        MainWindow.addDockWidget(QtCore.Qt.DockWidgetArea(1), self.eventOptionsDockWidget)
        self.stationOptionsDockWidget = QtGui.QDockWidget(MainWindow)
        self.stationOptionsDockWidget.setFloating(False)
        self.stationOptionsDockWidget.setFeatures(QtGui.QDockWidget.AllDockWidgetFeatures)
        self.stationOptionsDockWidget.setAllowedAreas(QtCore.Qt.LeftDockWidgetArea|QtCore.Qt.RightDockWidgetArea)
        self.stationOptionsDockWidget.setObjectName(_fromUtf8("stationOptionsDockWidget"))
        self.dockWidgetContents_2 = QtGui.QWidget()
        self.dockWidgetContents_2.setObjectName(_fromUtf8("dockWidgetContents_2"))
        self.stationOptionsDockWidget.setWidget(self.dockWidgetContents_2)
        MainWindow.addDockWidget(QtCore.Qt.DockWidgetArea(2), self.stationOptionsDockWidget)
        self.statusBar = QtGui.QStatusBar(MainWindow)
        self.statusBar.setObjectName(_fromUtf8("statusBar"))
        MainWindow.setStatusBar(self.statusBar)

        self.retranslateUi(MainWindow)
        QtCore.QMetaObject.connectSlotsByName(MainWindow)
        MainWindow.setTabOrder(self.getEventsButton, self.getStationsButton)
        MainWindow.setTabOrder(self.getStationsButton, self.eventsTable)
        MainWindow.setTabOrder(self.eventsTable, self.stationsTable)
        MainWindow.setTabOrder(self.stationsTable, self.getWaveformsButton)
        MainWindow.setTabOrder(self.getWaveformsButton, self.clearEventSelectionButton)
        MainWindow.setTabOrder(self.clearEventSelectionButton, self.clearStationSelectionButton)
        MainWindow.setTabOrder(self.clearStationSelectionButton, self.toggleEventOptions)
        MainWindow.setTabOrder(self.toggleEventOptions, self.toggleStationOptions)
        MainWindow.setTabOrder(self.toggleStationOptions, self.mapZoomInButton)
        MainWindow.setTabOrder(self.mapZoomInButton, self.mapZoomOutButton)
        MainWindow.setTabOrder(self.mapZoomOutButton, self.mapResetButton)

    def retranslateUi(self, MainWindow):
        MainWindow.setWindowTitle(_translate("MainWindow", "MainWindow", None))
        self.mapZoomInButton.setText(_translate("MainWindow", "Zoom In", None))
        self.mapZoomOutButton.setText(_translate("MainWindow", "Zoom Out", None))
        self.mapResetButton.setText(_translate("MainWindow", "Reset", None))
        self.toggleEventOptions.setText(_translate("MainWindow", "< Options", None))
        self.getEventsButton.setText(_translate("MainWindow", "Get Events", None))
        self.eventsTable.setSortingEnabled(True)
        self.clearEventSelectionButton.setText(_translate("MainWindow", "Clear Selection", None))
        self.getStationsButton.setText(_translate("MainWindow", "Get Stations", None))
        self.toggleStationOptions.setText(_translate("MainWindow", "Options >", None))
        self.stationsTable.setSortingEnabled(True)
        self.clearStationSelectionButton.setText(_translate("MainWindow", "Clear Selection", None))
        self.getWaveformsButton.setText(_translate("MainWindow", "Get Waveforms", None))
        self.eventOptionsDockWidget.setWindowTitle(_translate("MainWindow", "Event Options", None))
        self.stationOptionsDockWidget.setWindowTitle(_translate("MainWindow", "Station Options", None))

from pyweed.gui.uic.MyQt4MplCanvas import MyQt4MplCanvas



# -*- coding: utf-8 -*-
"""
Code for saving and loading user-selected events/stations.

:copyright:
    Mazama Science, IRIS
:license:
    GNU Lesser General Public License, Version 3
    (http://www.gnu.org/copyleft/lesser.html)
"""

from __future__ import (absolute_import, division, print_function)
import copy
from obspy.core.inventory.inventory import Inventory
from obspy.core.event.catalog import Catalog


def get_filtered_inventory(inventory, stations_iter):
    """
    Given an inventory and an iterator of selected network/station/channel items, return
    an inventory containing only the selected items.
    """
    networks = {}
    stations = {}
    for (network, station, channel) in stations_iter:
        # Create a station record if necessary, and add this channel to it
        full_station_code = "%s.%s" % (network.code, station.code)
        if full_station_code not in stations:
            f_station = copy.copy(station)
            f_station.channels = []
            stations[full_station_code] = f_station
            # Create a network record if necessary, and add this station to it
            if network.code not in networks:
                f_network = copy.copy(network)
                f_network.stations = []
                networks[network.code] = f_network
            networks[network.code].stations.append(f_station)
        stations[full_station_code].channels.append(channel)

    return Inventory(list(networks.values()), inventory.source, inventory.sender)


def get_filtered_catalog(catalog, events_iter):
    """
    Given a catalog and an iterator of selected events, return a catalog containing only the selected events.
    """
    return Catalog(list(events_iter))





# -*- coding: utf-8 -*-
"""
Main window

:copyright:
    Mazama Science, IRIS
:license:
    GNU Lesser General Public License, Version 3
    (http://www.gnu.org/copyleft/lesser.html)
"""

from __future__ import (absolute_import, division, print_function)

from pyweed import __version__, __app_name__

from PyQt4 import QtGui
from pyweed.gui.uic import MainWindow
from pyweed.preferences import safe_int, safe_bool, bool_to_str
import logging
from pyweed.pyweed_utils import iter_channels, get_preferred_origin, get_preferred_magnitude
from pyweed.gui.Seismap import Seismap
from pyweed.gui.EventOptionsWidget import EventOptionsWidget
from pyweed.gui.StationOptionsWidget import StationOptionsWidget
from pyweed.gui.TableItems import TableItems, Column
from pyweed.event_options import EventOptions
from PyQt4.QtCore import pyqtSlot
from pyweed.gui.SpinnerWidget import SpinnerWidget
from pyweed.station_options import StationOptions

LOGGER = logging.getLogger(__name__)


def make_exclusive(widget1, widget2):
    """
    Make two widgets exclusive, so that if one is toggled the other is disabled and vice versa
    """
    widget1.toggled.connect(
        lambda b: widget2.setEnabled(not b))
    widget2.toggled.connect(
        lambda b: widget1.setEnabled(not b))
    widget2.setEnabled(not widget1.isChecked())
    widget1.setEnabled(not widget2.isChecked())


class EventTableItems(TableItems):
    """
    Defines the table for displaying events
    """
    columns = [
        Column('Id'),
        Column('Date/Time'),
        Column('Magnitude'),
        Column('Longitude'),
        Column('Latitude'),
        Column('Depth'),
        Column('Location'),
    ]

    def rows(self, data):
        """
        Turn the data into rows (an iterable of lists) of QTableWidgetItems
        """
        for i, event in enumerate(data):
            origin = get_preferred_origin(event)
            if not origin:
                continue
            magnitude = get_preferred_magnitude(event)
            if not magnitude:
                continue
            time = origin.time.strftime("%Y-%m-%d %H:%M:%S")  # use strftime to remove milliseconds
            event_description = "no description"
            if len(event.event_descriptions) > 0:
                event_description = event.event_descriptions[0].text

            yield [
                self.numericWidget(i),
                self.stringWidget(time),
                self.numericWidget(magnitude.mag, "%s %s" % (magnitude.mag, magnitude.magnitude_type)),
                self.numericWidget(origin.longitude, "%.03f°"),
                self.numericWidget(origin.latitude, "%.03f°"),
                self.numericWidget(origin.depth / 1000, "%.02f km"),  # we wish to report in km
                self.stringWidget(event_description.title()),
            ]


class StationTableItems(TableItems):
    """
    Defines the table for displaying stations
    """
    columns = [
        Column('SNCL'),
        Column('Network'),
        Column('Station'),
        Column('Location'),
        Column('Channel'),
        Column('Longitude'),
        Column('Latitude'),
        Column('Description'),
    ]

    def rows(self, data):
        """
        Turn the data into rows (an iterable of lists) of QTableWidgetItems
        """
        sncls = set()
        for (network, station, channel) in iter_channels(data):
            sncl = '.'.join((network.code, station.code, channel.location_code, channel.code))
            if sncl in sncls:
                LOGGER.debug("Found duplicate SNCL: %s", sncl)
            else:
                sncls.add(sncl)
                yield [
                    self.stringWidget(sncl),
                    self.stringWidget(network.code),
                    self.stringWidget(station.code),
                    self.stringWidget(channel.location_code),
                    self.stringWidget(channel.code),
                    self.numericWidget(channel.longitude, "%.03f°"),
                    self.numericWidget(channel.latitude, "%.03f°"),
                    self.stringWidget(station.site.name),
                ]


class MainWindow(QtGui.QMainWindow, MainWindow.Ui_MainWindow):

    eventTableItems = None
    stationTableItems = None
    eventsSpinner = None
    stationsSpinner = None

    def __init__(self, pyweed):
        super(MainWindow, self).__init__()
        self.setupUi(self)
        # Note that at this point, pyweed is only minimally initialized!
        self.pyweed = pyweed

    def initialize(self):
        """
        Most of the initialization should happen here. PyWEED has to create the MainWindow first, before
        it's fully configured everything (mainly because it needs to attach various other widgets and things
        to MainWindow). So __init__() above needs to be very minimal. Once everything is ready, the application
        will call initialize() and we can safely do all the "real" initialization.
        """
        prefs = self.pyweed.preferences

        # Set MainWindow properties
        self.setWindowTitle('%s version %s' % (__app_name__, __version__))

        # Options widgets
        self.eventOptionsWidget = EventOptionsWidget(
            self, self.pyweed.event_options, self.pyweed.station_options,
            self.eventOptionsDockWidget, self.toggleEventOptions)
        self.stationOptionsWidget = StationOptionsWidget(
            self, self.pyweed.station_options, self.pyweed.event_options,
            self.stationOptionsDockWidget, self.toggleStationOptions)

        # When any options change, enable the relevant download button
        self.eventOptionsWidget.changed.connect(lambda: self.getEventsButton.setEnabled(True))
        self.stationOptionsWidget.changed.connect(lambda: self.getStationsButton.setEnabled(True))

        # Map
        self.initializeMap()

        # When the options coordinates change, we want to update the map
        self.eventOptionsWidget.changedCoords.connect(
            lambda: self.updateSeismap(events=True)
        )
        self.stationOptionsWidget.changedCoords.connect(
            lambda: self.updateSeismap(stations=True)
        )

        # Table selection
        self.eventsTable.itemSelectionChanged.connect(self.onEventSelectionChanged)
        self.stationsTable.itemSelectionChanged.connect(self.onStationSelectionChanged)
        self.clearEventSelectionButton.clicked.connect(self.eventsTable.clearSelection)
        self.clearStationSelectionButton.clicked.connect(self.stationsTable.clearSelection)

        # Main window buttons
        self.getEventsButton.clicked.connect(self.getEvents)
        self.getStationsButton.clicked.connect(self.getStations)
        self.getWaveformsButton.clicked.connect(self.getWaveforms)

        # Size and placement according to preferences
        self.resize(
            safe_int(prefs.MainWindow.width, 1000),
            safe_int(prefs.MainWindow.height, 800))
        self.eventOptionsDockWidget.setFloating(safe_bool(prefs.MainWindow.eventOptionsFloat, False))
        self.stationOptionsDockWidget.setFloating(safe_bool(prefs.MainWindow.stationOptionsFloat, False))

        # Add spinner overlays to the event/station widgets
        self.eventsSpinner = SpinnerWidget("Loading events...", parent=self.eventsWidget)
        self.eventsSpinner.cancelled.connect(self.pyweed.events_handler.cancel)
        self.stationsSpinner = SpinnerWidget("Loading stations...", parent=self.stationsWidget)
        self.stationsSpinner.cancelled.connect(self.pyweed.stations_handler.cancel)

        # Do an initial update
        self.updateSeismap(events=True)
        self.updateSeismap(stations=True)

        self.manageGetWaveformsButton()

    def initializeMap(self):
        LOGGER.info('Setting up main map...')

        self.seismap = Seismap(self.mapCanvas)

        # Map of buttons to the relevant draw modes
        self.drawButtons = {
            'events.box': self.eventOptionsWidget.drawLocationRangeToolButton,
            'events.toroid': self.eventOptionsWidget.drawDistanceFromPointToolButton,
            'stations.box': self.stationOptionsWidget.drawLocationRangeToolButton,
            'stations.toroid': self.stationOptionsWidget.drawDistanceFromPointToolButton,
        }

        # Generate a handler to toggle a given drawing mode
        def drawModeFn(mode):
            return lambda checked: self.seismap.toggleDrawMode(mode, checked)
        # Register draw mode toggle handlers
        for mode, button in self.drawButtons.items():
            button.clicked.connect(drawModeFn(mode))

        self.mapZoomInButton.clicked.connect(self.seismap.zoomIn)
        self.mapZoomOutButton.clicked.connect(self.seismap.zoomOut)
        self.mapResetButton.clicked.connect(self.seismap.zoomReset)

        self.seismap.drawEnd.connect(self.onMapDrawFinished)

    def showMessage(self, message):
        """ Display the given message in the status bar """
        self.statusBar.showMessage(message, 4000)

    @pyqtSlot(object)
    def onMapDrawFinished(self, event):
        """
        Called when Seismap emits a DrawEvent upon completion
        """
        # If the user finished an operation associated with a button, uncheck it
        button = self.drawButtons.get(event.mode)
        if button:
            button.setChecked(False)

        # For box/toroid bounds drawing, handle the returned geo data
        options = {}
        if event.points:
            if 'box' in event.mode:
                (n, e, s, w) = event.points
                options = {
                    'location_choice': EventOptions.LOCATION_BOX,  # Assumes StationOptions.LOCATION_BOX is the same
                    'maxlatitude': n,
                    'maxlongitude': e,
                    'minlatitude': s,
                    'minlongitude': w,
                }
            elif 'toroid' in event.mode:
                (lat, lon, dist) = event.points
                options = {
                    'location_choice': EventOptions.LOCATION_POINT,  # Assumes StationOptions.LOCATION_POINT is the same
                    'latitude': lat,
                    'longitude': lon,
                    'maxradius': dist
                }
            # Set event or station options
            if 'events' in event.mode:
                self.pyweed.set_event_options(options)
            elif 'stations' in event.mode:
                self.pyweed.set_station_options(options)

    def getEvents(self):
        """
        Trigger the event retrieval from web services
        """
        LOGGER.info('Loading events...')
        self.eventsSpinner.show()
        self.getEventsButton.setEnabled(False)
        self.pyweed.fetch_events()

    def updateSeismap(self, events=False, stations=False):
        """
        Update seismap when [event|station]OptionsWidget coordinates change
        """
        LOGGER.debug("Updating map from widget: %s" % ((events and "events") or "stations"))

        if events:
            options = self.eventOptionsWidget.getOptions()
            markers = self.seismap.eventMarkers
        else:
            options = self.stationOptionsWidget.getOptions()
            markers = self.seismap.stationMarkers

        try:
            self.seismap.clearBoundingMarkers(markers)
            if options["location_choice"] == EventOptions.LOCATION_BOX:
                self.seismap.addMarkerBox(
                    markers,
                    float(options["maxlatitude"]),
                    float(options["maxlongitude"]),
                    float(options["minlatitude"]),
                    float(options["minlongitude"])
                )
            elif options["location_choice"] == EventOptions.LOCATION_POINT:
                self.seismap.addMarkerToroid(
                    markers,
                    float(options["latitude"]),
                    float(options["longitude"]),
                    float(options["minradius"]),
                    float(options["maxradius"])
                )
            elif options["location_choice"] == StationOptions.LOCATION_EVENTS:
                # Show distance markers around all events
                for event in self.pyweed.iter_selected_events():
                    origin = get_preferred_origin(event)
                    if origin:
                        self.seismap.addMarkerToroid(
                            markers,
                            origin.latitude,
                            origin.longitude,
                            float(options["mindistance"]),
                            float(options["maxdistance"])
                        )
        except Exception as e:
            LOGGER.error("Failed to update seismap! %s", e, exc_info=True)

    def onEventsLoaded(self, events):
        """
        Handler triggered when the EventsHandler finishes loading events
        """
        self.eventsSpinner.hide()

        if isinstance(events, Exception):
            self.eventSelectionLabel.setText('Error! See log for details')
            msg = "Error loading events: %s" % events
            LOGGER.error(msg)
            self.showMessage(msg)
            return

        if not self.eventTableItems:
            self.eventTableItems = EventTableItems(self.eventsTable)
        self.eventTableItems.fill(events)

        # Add items to the map -------------------------------------------------

        self.seismap.addEvents(events)

        self.onEventSelectionChanged()
        status = 'Finished loading events'
        LOGGER.info(status)
        self.showMessage(status)

    def getStations(self):
        """
        Trigger the channel metadata retrieval from web services
        """
        LOGGER.info('Loading stations...')
        self.stationsSpinner.show()
        self.getStationsButton.setEnabled(False)
        self.pyweed.fetch_stations()

    def onStationsLoaded(self, stations):
        """
        Handler triggered when the StationsHandler finishes loading stations
        """
        self.stationsSpinner.hide()

        if isinstance(stations, Exception):
            self.stationSelectionLabel.setText('Error! See log for details')
            msg = "Error loading stations: %s" % stations
            LOGGER.error(msg)
            self.showMessage(msg)
            return

        if not self.stationTableItems:
            self.stationTableItems = StationTableItems(self.stationsTable)
        self.stationTableItems.fill(stations)

        # Add items to the map -------------------------------------------------

        self.seismap.addStations(stations)

        self.onStationSelectionChanged()
        status = 'Finished loading stations'
        LOGGER.info(status)
        self.showMessage(status)

    def getWaveforms(self):
        self.pyweed.openWaveformsDialog()

    def selectAllEvents(self):
        """
        Select all events in the table. This is mainly for loading data from a summary file, where
        we want all the data to be pre-selected.
        """
        self.eventsTable.selectAll()

    def onEventSelectionChanged(self):
        """
        Handle a click anywhere in the table.
        """
        # Get selected ids
        ids = []
        for idx in self.eventsTable.selectionModel().selectedRows():
            ids.append(int(self.eventsTable.item(idx.row(), 0).text()))

        numSelected = len(ids)
        numTotal = self.eventsTable.rowCount()
        self.eventSelectionLabel.setText(
            "Selected %d of %d events" % (numSelected, numTotal))

        # Get locations and event IDs
        points = []
        eventIDs = []
        for id in ids:
            event = self.pyweed.events[id]
            origin = get_preferred_origin(event)
            if not origin:
                continue
            points.append((origin.latitude, origin.longitude))
            eventIDs.append(event.resource_id.id)

        # Update the events_handler with the latest selection information
        self.pyweed.set_selected_event_ids(eventIDs)

        self.seismap.addEventsHighlighting(points)

        self.stationOptionsWidget.onEventSelectionChanged()

        self.manageGetWaveformsButton()

    def selectAllStations(self):
        """
        Select all stations in the table. This is mainly for loading data from a summary file, where
        we want all the data to be pre-selected.
        """
        self.stationsTable.selectAll()

    def onStationSelectionChanged(self):
        # Get selected sncls
        sncls = []
        for idx in self.stationsTable.selectionModel().selectedRows():
            sncls.append(self.stationsTable.item(idx.row(), 0).text())

        numSelected = len(sncls)
        numTotal = self.stationsTable.rowCount()
        self.stationSelectionLabel.setText(
            "Selected %d of %d channels" % (numSelected, numTotal))

        # Get locations
        points = []
        for sncl in sncls:
            try:
                coordinates = self.pyweed.stations.get_coordinates(sncl)
                points.append((coordinates['latitude'], coordinates['longitude']))
            except:
                pass

        # Update the stations_handler with the latest selection information
        self.pyweed.set_selected_station_ids(sncls)

        self.seismap.addStationsHighlighting(points)

        self.manageGetWaveformsButton()

    def manageGetWaveformsButton(self):
        """
        Handle enabled/disabled status of the "Get Waveforms" button
        based on the presence/absence of selected events and stations
        """
        if self.pyweed.selected_event_ids and self.pyweed.selected_station_ids:
            self.getWaveformsButton.setEnabled(True)
        else:
            self.getWaveformsButton.setEnabled(False)

    def closeEvent(self, event):
        """
        When the main window closes, quit the application
        """
        self.pyweed.closeApplication()
        event.ignore()

    def savePreferences(self):
        prefs = self.pyweed.preferences

        prefs.MainWindow.width = self.width()
        prefs.MainWindow.height = self.height()
        prefs.MainWindow.eventOptionsFloat = bool_to_str(self.eventOptionsDockWidget.isFloating())
        prefs.MainWindow.stationOptionsFloat = bool_to_str(self.stationOptionsDockWidget.isFloating())



# -*- coding: utf-8 -*-
"""
Base classes for EventOptionsWidet and StationOptionsWidget

:copyright:
    Mazama Science, IRIS
:license:
    GNU Lesser General Public License, Version 3
    (http://www.gnu.org/copyleft/lesser.html)
"""

from PyQt4 import QtGui, QtCore
import logging
from distutils.util import strtobool

LOGGER = logging.getLogger(__name__)


class BaseOptionsWidget(QtGui.QDialog):
    """
    Base functionality for the EventOptionsWidget and StationOptionsWidget.
    """

    # We want to watch for changes in the widget inputs, but every type of input has a different one so
    # we need to try a bunch of options
    INPUT_CHANGE_SIGNALS = ('valueChanged', 'textChanged', 'dateTimeChanged', 'clicked',)

    # Signal to indicate that the options have changed
    changed = QtCore.pyqtSignal(object)
    # Signal indicating that the user changed a location parameter
    changedCoords = QtCore.pyqtSignal(object)
    # Map of the individual inputs by name
    inputs = None

    def __init__(self, parent, options, otherOptions, dockWidget, toggleButton):
        super(BaseOptionsWidget, self).__init__(parent=parent)
        # This function just saves variables and do some low-level widget operations
        # then calls `initialize()` for the higher-level behavior.
        self.setupUi(self)
        self.inputs = self.mapInputs()
        self.options = options
        self.otherOptions = otherOptions

        # Add to dock
        dockWidget.setWidget(self)
        # Connect the toggle switch for hiding the dock
        toggleButton.toggled.connect(dockWidget.setVisible)
        dockWidget.visibilityChanged.connect(toggleButton.setChecked)

        # Call a separate function for the higher-level behavior, so subclasses can extend
        self.initialize()

    def initialize(self):
        """
        Second part of the initialization, subclass initialization should extend this.
        """
        # Push values to the inputs before connecting them
        self.setOptions()
        self.connectInputs()

        # Hook up the shortcut buttons
        self.time30DaysPushButton.clicked.connect(self.setTime30Days)
        self.time1YearPushButton.clicked.connect(self.setTime1Year)

        # Hook up the copy buttons
        self.get_timeFromOtherButton().clicked.connect(self.copyTimeOptions)
        self.get_locationFromOtherButton().clicked.connect(self.copyLocationOptions)

    def mapInputs(self):
        """
        Subclasses override this to provide a map of keys to input widgets

        ex:
        return {
            'network': self.networkLineEdit,
            'station': self.stationLineEdit,
            'location': self.locationLineEdit,
            'channel': self.channelLineEdit,
            ...
        }
        """
        raise NotImplementedError()

    def get_timeFromOtherButton(self):
        """ Return the button that triggers copying time from the other widget """
        raise NotImplementedError()

    def get_locationFromOtherButton(self):
        """ Return the button that triggers copying location from the other widget """
        raise NotImplementedError()

    def connectInput(self, key, one_input):
        # Different widgets emit different signals  :P
        for signal_name in self.INPUT_CHANGE_SIGNALS:
            if hasattr(one_input, signal_name):
                LOGGER.debug("Listening to %s.%s" % (one_input, signal_name))
                getattr(one_input, signal_name).connect(lambda: self.onInputChanged(key))
                return

    def connectInputs(self):
        """
        Set up listeners for the inputs
        """
        for key, one_input in self.inputs.items():
            self.connectInput(key, one_input)

    def onInputChanged(self, key):
        """
        Called when any input is changed
        """
        LOGGER.debug("Input changed: %s" % key)
        # Update the core options object
        self.options.set_options(self.getOptions())
        # Emit a change event
        self.changed.emit(key)
        # Emit a coordinate change event if appropriate
        if self.isCoordinateInput(key):
            self.changedCoords.emit(key)

    def isCoordinateInput(self, key):
        """
        Return true if the given key represents a coordinate input.
        Subclasses may override/extend this.
        """
        for marker in ('latitude', 'longitude', 'radius', '_location', 'distance',):
            if marker in key:
                return True
        return False

    def getInputValue(self, key):
        """
        Get the value of a single input. This is intended to Pythonify the Qt values that come natively
        from the widgets. (ie. return a Python date rather than a QDate).

        This is used internally by getOptions()
        """
        one_input = self.inputs.get(key)
        if one_input:
            if isinstance(one_input, QtGui.QDateTimeEdit):
                # DateTime
                return one_input.dateTime().toString(QtCore.Qt.ISODate)
            elif isinstance(one_input, QtGui.QAbstractButton):
                # Radio/checkbox button
                return str(one_input.isChecked())
            elif isinstance(one_input, QtGui.QComboBox):
                return one_input.currentText()
            elif hasattr(one_input, 'text'):
                return one_input.text()
            raise Exception("Couldn't identify the QWidget type for %s (%s)" % (key, one_input))
        return None

    def setInputValue(self, key, value):
        """
        Set the input value based on a string from the options
        """
        one_input = self.inputs.get(key)
        if one_input:
            if isinstance(one_input, QtGui.QDateTimeEdit):
                # Ugh, complicated conversion from UTCDateTime
                dt = QtCore.QDateTime.fromString(value, QtCore.Qt.ISODate)
                one_input.setDateTime(dt)
            elif isinstance(one_input, QtGui.QDoubleSpinBox):
                # Float value
                one_input.setValue(float(value))
            elif isinstance(one_input, QtGui.QComboBox):
                # Combo box
                index = one_input.findText(value)
                if index > -1:
                    one_input.setCurrentIndex(index)
            elif isinstance(one_input, QtGui.QLineEdit):
                # Text input
                one_input.setText(value)
            elif isinstance(one_input, QtGui.QAbstractButton):
                # Radio/checkbox button
                one_input.setChecked(strtobool(str(value)))
            else:
                raise Exception("Don't know how to set an input for %s (%s)" % (key, one_input))

    def setOptions(self):
        """
        Put the current set of options into the mapped inputs
        """
        # Get a dictionary of stringified options values
        for key, value in self.optionsToInputs(self.options.get_options(stringify=True)).items():
            try:
                self.setInputValue(key, value)
            except Exception as e:
                LOGGER.warning("Unable to set input value for %s: %s", key, e)

    def optionsToInputs(self, values):
        """
        Hook for subclasses to modify a dictionary of options to map to the input widgets
        """
        return values

    def getOptions(self):
        """
        Return a dictionary containing the input values, formatted as appropriate for a service call.
        """
        inputValues = {}
        for key in self.inputs.keys():
            try:
                inputValues[key] = self.getInputValue(key)
            except Exception as e:
                LOGGER.warning("Unable to get input value %s: %s", key, e)
        return self.inputsToOptions(inputValues)

    def inputsToOptions(self, values):
        """
        Hook for subclasses to modify a dictionary of input values to map to the options
        """
        return values

    def copyTimeOptions(self):
        """
        Copy the time options from event_options/station_options
        """
        LOGGER.info("Copying time to %s" % self.__class__)
        time_options = self.otherOptions.get_time_options()
        time_options.update(self.otherOptions.get_options(['time_choice']))
        self.options.set_options(time_options)
        self.setOptions()
        self.changed.emit('time_choice')

    def copyLocationOptions(self):
        """
        Copy the location options from event_options/station_options
        """
        LOGGER.info("Copying location to %s" % self.__class__)
        loc_options = self.otherOptions.get_location_options()
        loc_options.update(self.otherOptions.get_options(['location_choice']))
        self.options.set_options(loc_options)
        self.setOptions()
        self.changed.emit('location_choice')
        self.changedCoords.emit('location_choice')

    def setTime30Days(self):
        """
        Set the start/end times to cover the last 30 days
        """
        end = QtCore.QDateTime.currentDateTimeUtc()
        start = end.addDays(-30)
        self.starttimeDateTimeEdit.setDateTime(start)
        self.endtimeDateTimeEdit.setDateTime(end)
        self.changed.emit('time_choice')

    def setTime1Year(self):
        """
        Set the start/end times to cover the last 30 days
        """
        end = QtCore.QDateTime.currentDateTimeUtc()
        start = end.addYears(-1)
        self.starttimeDateTimeEdit.setDateTime(start)
        self.endtimeDateTimeEdit.setDateTime(end)
        self.changed.emit('time_choice')





# -*- coding: utf-8 -*-
"""
Container for stations.

:copyright:
    Mazama Science, IRIS
:license:
    GNU Lesser General Public License, Version 3
    (http://www.gnu.org/copyleft/lesser.html)
"""

from __future__ import (absolute_import, division, print_function)

import logging
from pyweed.signals import SignalingThread, SignalingObject
from obspy.core.inventory.inventory import Inventory
from pyweed.pyweed_utils import get_service_url, CancelledException, DataRequest, get_distance
from PyQt4 import QtCore
import concurrent.futures
from pyweed.dist_from_events import get_combined_locations, CrossBorderException

LOGGER = logging.getLogger(__name__)


def load_stations(client, parameters):
    """
    Execute one query for station metadata. This is a standalone function so we can
    run it in a separate thread.
    """
    try:
        LOGGER.info('Loading stations: %s', get_service_url(client, 'station', parameters))
        return client.get_stations(**parameters)
    except Exception as e:
        # If no results found, the client will raise an exception, we need to trap this
        # TODO: this should be much cleaner with a fix to https://github.com/obspy/obspy/issues/1656
        if str(e).startswith("No data"):
            LOGGER.warning("No stations found! Your query may be too narrow.")
            return Inventory([], 'INTERNAL')
        else:
            raise


class StationsLoader(SignalingThread):
    """
    Thread to handle station requests
    """
    progress = QtCore.pyqtSignal()

    def __init__(self, request):
        """
        Initialization.
        """
        # Keep a reference to globally shared components
        self.request = request
        self.futures = {}
        super(StationsLoader, self).__init__()

    def run(self):
        """
        Make a webservice request for events using the passed in options.
        """
        self.setPriority(QtCore.QThread.LowestPriority)
        self.clearFutures()
        self.futures = {}

        inventory = None
        LOGGER.info("Making %d station requests" % len(self.request.sub_requests))
        with concurrent.futures.ThreadPoolExecutor(5) as executor:
            for sub_request in self.request.sub_requests:
                # Dictionary lets us look up argument by result later
                self.futures[executor.submit(load_stations, self.request.client, sub_request)] = sub_request
            # Iterate through Futures as they complete
            for result in concurrent.futures.as_completed(self.futures):
                LOGGER.debug("Stations loaded")
                try:
                    if not inventory:
                        inventory = result.result()
                    else:
                        inventory += result.result()
                    self.progress.emit()
                except Exception:
                    self.progress.emit()
        self.futures = {}
        # If no inventory object (ie. no sub-requests were run) create a dummy one
        if not inventory:
            inventory = Inventory([], 'INTERNAL')
        inventory = self.request.process_result(inventory)
        self.done.emit(inventory)

    def clearFutures(self):
        """
        Cancel any outstanding tasks
        """
        if self.futures:
            for future in self.futures:
                if not future.done():
                    LOGGER.debug("Cancelling unexecuted future")
                    future.cancel()

    def cancel(self):
        """
        User-requested cancel
        """
        self.done.disconnect()
        self.progress.disconnect()
        self.clearFutures()


class StationsDataRequest(DataRequest):
    event_locations = None
    distance_range = None

    def __init__(self, client, base_options, distance_range, event_locations):
        """
        :param client: an ObsPy FDSN client
        :param base_options: the basic query options (ie. from StationOptions)
        :param distance_range: a tuple of (min, max) if querying by distance from events
        :param event_locations: a list of selected event locations if querying by distance from events
        """
        super(StationsDataRequest, self).__init__(client)
        if distance_range and event_locations:
            # Get a list of just the (lat, lon) for each event
            self.event_locations = list((loc[1] for loc in event_locations))
            self.distance_range = distance_range
            try:
                combined_locations = get_combined_locations(self.event_locations, self.distance_range['maxdistance'])
                self.sub_requests = [
                    dict(
                        base_options,
                        minlatitude=box.lat1,
                        maxlatitude=box.lat2,
                        minlongitude=box.lon1,
                        maxlongitude=box.lon2
                    )
                    for box in combined_locations
                ]
            except CrossBorderException:
                # Can't break into subqueries, return the base (global, we assume) query
                LOGGER.warning("Couldn't calculate a bounding box for events, using global")
                self.sub_requests = [base_options]
        else:
            self.sub_requests = [base_options]

    def filter_one_station(self, station):
        """
        Filter one station from the results
        """
        for lat, lon in self.event_locations:
            dist = get_distance(lat, lon, station.latitude, station.longitude)
            if self.distance_range['mindistance'] <= dist <= self.distance_range['maxdistance']:
                return True
        return False

    def process_result(self, result):
        """
        If the request is based on distance from a set of events, we need to perform
        a filter after the request, since the low-level request probably overselected.
        """
        result = super(StationsDataRequest, self).process_result(result)
        if isinstance(result, Inventory) and self.event_locations and self.distance_range:
            filtered_networks = []
            for network in result:
                filtered_stations = [station for station in network if self.filter_one_station(station)]
                if filtered_stations:
                    network.stations = filtered_stations
                    filtered_networks.append(network)
            result.networks = filtered_networks
        return result


class StationsHandler(SignalingObject):
    """
    Container for stations.
    """

    def __init__(self, pyweed):
        """
        Initialization.
        """
        super(StationsHandler, self).__init__()

        self.pyweed = pyweed
        self.inventory_loader = None

    def load_inventory(self, request):
        try:
            self.inventory_loader = StationsLoader(request)
            self.inventory_loader.done.connect(self.on_inventory_loaded)
            self.inventory_loader.start()
        except Exception as e:
            self.done.emit(e)

    def on_inventory_loaded(self, inventory):
        self.done.emit(inventory)

    def cancel(self):
        if self.inventory_loader:
            self.inventory_loader.done.disconnect()
        self.done.emit(CancelledException())


# ------------------------------------------------------------------------------
# Main
# ------------------------------------------------------------------------------

if __name__ == '__main__':
    import doctest
    doctest.testmod(exclude_empty=True)



from time import sleep
from pyweed.pyweed_utils import get_distance, get_arrivals, TimeWindow
import unittest
from obspy.core.utcdatetime import UTCDateTime


def gui_test(pyweed):
    pyweed.fetch_events()
    pyweed.fetch_stations()
    sleep(2)
    for i in range(5):
        pyweed.mainWindow.eventsTable.selectRow(i)
    for i in range(5):
        pyweed.mainWindow.stationsTable.selectRow(i)
    pyweed.openWaveformsDialog()


class DistanceTest(unittest.TestCase):
    """
    Some sanity tests, mainly to ensure we have the units right
    """
    def test_distance_1(self):
        dist = get_distance(0, 0, 10, 0)
        self.assertAlmostEqual(dist, 10, delta=1)

    def test_distance_2(self):
        dist = get_distance(0, 0, 0, 10)
        self.assertAlmostEqual(dist, 10, delta=1)

    def test_distance_3(self):
        dist = get_distance(0, 170, 0, -170)
        self.assertAlmostEqual(dist, 20, delta=1)


class ArrivalsTest(unittest.TestCase):
    def test_arrivals_1(self):
        arrivals = get_arrivals(20, 100)
        self.assertAlmostEqual(arrivals['P'], 264, delta=1)
        self.assertAlmostEqual(arrivals['S'], 484, delta=1)


class TimeWindowTest(unittest.TestCase):
    def test_timewindow_1(self):
        time_window = TimeWindow(60, 600, 'P', 'P')
        event_time = UTCDateTime("2017-01-01T12:00:00")
        arrivals = {
            'P': 264
        }
        (start, end) = time_window.calculate_window(
            event_time, arrivals)
        # 264s arrival time, -60s offset
        expected_start = UTCDateTime("2017-01-01T12:03:24")
        # 264s arrival time, 600s offset
        expected_end = UTCDateTime("2017-01-01T12:14:24")
        self.assertAlmostEqual(
            start, expected_start, delta=1)
        self.assertAlmostEqual(
            end, expected_end, delta=1)


if __name__ == '__main__':
    unittest.main()
